#!/usr/bin/python
#
# Block tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Unit testst for cs.vt.block.
'''

import sys
from random import choice
import unittest
from cs.binary_tests import _TestPacketFields
from .randutils import rand0, randblock
from . import totext, block as block_module
from .block import Block, \
    IndirectBlock, \
    RLEBlock, LiteralBlock, SubBlock, \
    verify_block, BlockType, \
    BlockRecord
from .store import MappingStore

class TestDataFilePacketFields(_TestPacketFields, unittest.TestCase):
  ''' Hook to test the hash PacketFields.
  '''

  def setUp(self):
    ''' Test the block module PacketField classes.
    '''
    self.module = block_module

class TestAll(unittest.TestCase):
  ''' All Block tests.
  '''

  def setUp(self):
    ''' Make a dict backed MappingStore.
    '''
    self.S = MappingStore("TestAll", {})

  def _verify_block(self, B, **kw):
    with self.subTest(task="_verify_block", block=B, **kw):
      BR = BlockRecord(B)
      BRbs = bytes(BR)
      BR2, offset = BlockRecord.from_bytes(BRbs)
      self.assertEqual(offset, len(BRbs))
      self.assertEqual(BR, BR2)
      for err in verify_block(B):
        raise ValueError("verify_block(%s): %s" % (B, err))
      errs = list(verify_block(B, **kw))
      self.assertEqual(errs, [])

  def _make_random_Block(self, block_type=None, size=None, leaf_only=False):
    with self.subTest(
        task="_make_random_Block",
        block_type=block_type, size=size, leaf_only=leaf_only
    ):
      if block_type is None:
        choices = [
            BlockType.BT_HASHCODE,
            BlockType.BT_RLE,
            BlockType.BT_LITERAL,
        ]
        if not leaf_only:
          choices.append(BlockType.BT_SUBBLOCK)
          choices.append(BlockType.BT_INDIRECT)
        block_type = choice(choices)
      if size is None:
        size = rand0(16385)
      with self.subTest(
          subtask="instantiate",
          block_type=block_type, size=size,
      ):
        if block_type == BlockType.BT_INDIRECT:
          subblocks = [self._make_random_Block() for _ in range(rand0(8))]
          B = IndirectBlock(subblocks, force=True)
        elif block_type == BlockType.BT_HASHCODE:
          rs = randblock(size)
          B = Block(data=rs)
          # we can get a literal block back - this is acceptable
          if B.type == BlockType.BT_LITERAL:
            block_type = BlockType.BT_LITERAL
        elif block_type == BlockType.BT_RLE:
          rb = bytes((rand0(256),))
          B = RLEBlock(size, rb)
        elif block_type == BlockType.BT_LITERAL:
          rs = randblock(size)
          B = LiteralBlock(data=rs)
        elif block_type == BlockType.BT_SUBBLOCK:
          B2 = self._make_random_Block()
          self._verify_block(B2)
          if len(B2) == 0:
            suboffset = 0
            subspan = 0
          else:
            suboffset = rand0(B2.span)
            subspan = rand0(B2.span - suboffset)
          B = SubBlock(B2, suboffset, subspan)
          # SubBlock returns an empty literal for an empty subblock
          if subspan == 0:
            block_type = BlockType.BT_LITERAL
        else:
          raise ValueError("unknow block type")
        self.assertEqual(
            B.type, block_type,
            "new Block is wrong type: %r, should be %r"
            % (B.type, block_type,))
        self._verify_block(B)
      return B

  def test00Block(self):
    ''' Make some random blocks, check size and content.
    '''
    with self.S:
      for _ in range(16):
        size = rand0(16385)
        rs = randblock(size)
        self.assertEqual(len(rs), size)
        B = Block(data=rs)
        self._verify_block(B)
        self.assertEqual(len(B), size)
        self.assertEqual(B.span, size)
        self.assertEqual(B.get_spanned_data(), rs)

  def test10IndirectBlock(self):
    ''' Construct various random indirect blocks and test.
    '''
    S = self.S
    with S:
      for _ in range(64):
        with self.subTest(loop=_):
          chunks = []
          subblocks = []
          total_length = 0
          for _ in range(rand0(16)):
            B = self._make_random_Block()
            subblocks.append(B)
            total_length += B.span
            chunks.append(B.get_spanned_data())
          fullblock = b''.join(chunks)
          IB = IndirectBlock(subblocks=subblocks, force=True)
          self._verify_block(IB, recurse=True)
          IBspan = IB.span
          self.assertEqual(
              IBspan, total_length,
              "IBspan(%d) != total_length(%d)" % (IB.span, total_length))
          IBH = IB.superblock.hashcode
          IBdata = IB.get_spanned_data()
          self.assertEqual(len(IBdata), total_length)
          self.assertEqual(IBdata, fullblock)
          # refetch block by hashcode
          IB2 = IndirectBlock(hashcode=IBH, span=len(IBdata))
          self._verify_block(IB2, recurse=True)
          IB2data = IB2.get_spanned_data()
          self.assertEqual(
              IBdata, IB2data,
              "IB:  %s\nIB2: %s" % (totext(IBdata), totext(IB2data)))
          for _ in range(32):
            with self.subTest(loop2=_):
              start = rand0(len(IB) + 1)
              length = rand0(len(IB) - start + 1) if start < len(IB) else 0
              end = start + length
              with self.subTest(start=start, end=end):
                chunk1 = IB[start:end]
                self.assertEqual(len(chunk1), length)
                chunk1a = fullblock[start:end]
                self.assertEqual(len(chunk1a), length)
                self.assertEqual(
                    chunk1, chunk1a,
                    "IB[%d:%d] != fullblock[%d:%d]" % (start, end, start, end))
                chunk2 = IB2[start:end]
                self.assertEqual(len(chunk2), length)
                self.assertEqual(
                    chunk1, chunk2,
                    "IB[%d:%d] != IB2[%d:%d]" % (start, end, start, end))

  def test02RoundTripSingleBlock(self):
    ''' Generate various block types, serialise then deserialise each.
    '''
    S = self.S
    with S:
      for block_type in BlockType.BT_HASHCODE, BlockType.BT_RLE, \
                        BlockType.BT_LITERAL, BlockType.BT_SUBBLOCK, \
                        BlockType.BT_INDIRECT:
        size = rand0(16385)
        with self.subTest(type=block_type, size=size):
          B = self._make_random_Block(block_type=block_type)
          Bserial = B.encode()
          B2, offset = BlockRecord.value_from_bytes(Bserial)
          self.assertEqual(
              offset, len(Bserial),
              "decoded %d bytes but len(Bserial)=%d" % (offset, len(Bserial)))
          self._verify_block(B2)
          if block_type != BlockType.BT_INDIRECT:
            self.assertEqual(B.type, B2.type, "block types differ")
            self.assertEqual(B.indirect, B2.indirect, "block indirects differ")
          self.assertEqual(B.span, B2.span, "span lengths differ")
          self.assertEqual(
                  B.get_spanned_data(),
                  B2.get_spanned_data(),
                  "spanned data differ")
          Btype = B2.type
          if Btype == BlockType.BT_INDIRECT:
            self.assertTrue(B.indirect)
            self._verify_block(B2.superblock)
          else:
            self.assertFalse(B.indirect)
            self.assertEqual(B.span, sum(len(chunk) for chunk in B.datafrom()))
            if Btype == BlockType.BT_HASHCODE:
              self.assertEqual(B.hashcode, B2.hashcode)
            elif Btype == BlockType.BT_RLE:
              self.assertEqual(B2.get_spanned_data(), B2.octet * B2.span)
            elif Btype == BlockType.BT_LITERAL:
              raise unittest.SkipTest("no specific test for LiteralBlock")
            elif Btype == BlockType.BT_SUBBLOCK:
              self._verify_block(B2.superblock)
            else:
              raise unittest.SkipTest(
                  "no type specific tests for Block type %r" % (block_type,))

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python -tt
#
# Data stores based on local files.
# - Cameron Simpson <cs@cskk.id.au>
#

''' DataDir: the sharable directory storing DataFiles used by DataDirStores.
'''

from binascii import hexlify
from collections import namedtuple
from collections.abc import Mapping
import errno
import os
from os import SEEK_SET, SEEK_CUR, SEEK_END
from os.path import (
    basename,
    isdir as isdirpath,
    isfile as isfilepath,
    join as joinpath,
    relpath)
import sqlite3
import stat
import sys
import time
from types import SimpleNamespace
from uuid import uuid4
from cs.app.flag import DummyFlags, FlaggedMixin
from cs.cache import LRU_Cache
from cs.excutils import logexc
from cs.fileutils import (
    DEFAULT_READSIZE,
    ReadMixin,
    datafrom_fd,
    makelockfile,
    read_from,
    shortpath,
    TimeoutError)
from cs.logutils import debug, info, warning, error, exception
from cs.pfx import Pfx, PfxThread as Thread, XP
from cs.py.func import prop as property
from cs.queues import IterableQueue
from cs.resources import MultiOpenMixin, RunStateMixin
from cs.seq import imerge
from cs.serialise import get_bs, put_bs
from cs.threads import locked
from cs.units import transcribe_bytes_geek
from . import MAX_FILE_SIZE, Lock, RLock
from .archive import Archive
from .block import Block
from .blockify import top_block_for, blocked_chunks_of, spliced_blocks, DEFAULT_SCAN_SIZE
from .datafile import DataFileReader, DataFileWriter, DATAFILE_DOT_EXT
from .dir import Dir, FileDirent
from .hash import (
    DEFAULT_HASHCLASS,
    HashCode, HashCodeUtilsMixin, MissingHashcodeError
)
from .index import choose as choose_indexclass
from .parsers import scanner_from_filename

DEFAULT_DATADIR_STATE_NAME = 'default'

# 1GiB rollover
DEFAULT_ROLLOVER = MAX_FILE_SIZE

# flush the index after this many updates in the index updater worker thread
INDEX_FLUSH_RATE = 16384

class DataFileState(SimpleNamespace):
  ''' General state information about a data file in use by a files based data dir.

      Attributes:
      * `datadir`: the _FilesDir tracking this state.
      * `filenum`: the numeric index of this file.
      * `filename`: path relative to the _FilesDir's data directory.
      * `indexed_to`: the maximum amount of data scanned and indexed
        so far.
  '''

  def __init__(
      self,
      datadir, filenum, filename,
      indexed_to=0, scanned_to=None,
  ) -> None:
    if scanned_to is None:
      scanned_to = indexed_to
    self.datadir = datadir
    self.filenum = filenum
    self.filename = filename
    self.indexed_to = indexed_to
    self.scanned_to = scanned_to

  def __str__(self):
    return "%s(%d:%r)" % (type(self).__name__, self.filenum, self.filename)

  @property
  def pathname(self):
    ''' Return the full pathname of this data file.
    '''
    return self.datadir.datapathto(self.filename)

  def stat_size(self, follow_symlinks=False):
    ''' Stat the datafile, return its size.
    '''
    path = self.pathname
    if follow_symlinks:
      S = os.stat(path)
    else:
      S = os.lstat(path)
    if not stat.S_ISREG(S.st_mode):
      return None
    return S.st_size

  def scanfrom(self, offset=0, **kw):
    ''' Scan this datafile from the supplied `offset` (default 0)
        yielding (offset, flags, data, post_offset).

        We use the DataDir's .scanfrom method because it knows the
        format of the file.
    '''
    yield from self.datadir.scanfrom(self.pathname, offset=offset, **kw)

class _FilesDir(HashCodeUtilsMixin, MultiOpenMixin, RunStateMixin, FlaggedMixin, Mapping):
  ''' Base class for locally stored data in files.

      There are two main subclasses of this at present:

      DataDir, where the data are kept in a subdirectory of UUID-named
      files, supporting easy merging and updating.

      PlatonicDataDir, where the data are present in a normal file tree,
      such as a preexisting media server directory or the like.
  '''

  STATE_FILENAME_FORMAT = 'index-{hashname}-state.sqlite'
  INDEX_FILENAME_BASE_FORMAT = 'index-{hashname}'
  DATA_SUBDIR = 'data'

  def __init__(
      self,
      statedirpath, hashclass,
      *,
      indexclass=None,
      flags=None, flag_prefix=None,
  ):
    ''' Initialise the DataDir with `statedirpath` and `datadirpath`.

        Parameters:
        * `statedirpath`: a directory containing state information about the
          DataFiles; this is the index-state.csv file and the associated
          index dbm-ish files.
        * `hashclass`: the hash class used to index chunk contents.
        * `indexclass`: the IndexClass providing the index to chunks in the
          DataFiles. If not specified, a supported index class with an
          existing index file will be chosen, otherwise the most favoured
          indexclass available will be chosen.
        * `flags`: optional Flags object for control; if specified then
          `flag_prefix` is also required
        * `flag_prefix`: prefix for control flag names

        Note that __init__ only saves the settings such as the `indexclass`
        and ensures that requisite directories exist.
        The monitor thread and runtime state are setup by the `startup` method
        and closed down by the `shutdown` method.
    '''
    assert isinstance(statedirpath, str)
    assert issubclass(hashclass, HashCode)
    RunStateMixin.__init__(self)
    MultiOpenMixin.__init__(self, lock=RLock())
    if flags is None:
      if flag_prefix is None:
        flags = DummyFlags()
        flag_prefix = 'DUMMY'
    else:
      if flag_prefix is None:
        raise ValueError("flags provided but no flag_prefix")
    FlaggedMixin.__init__(self, flags=flags, prefix=flag_prefix)
    self.statedirpath = statedirpath
    if hashclass is None:
      hashclass = DEFAULT_HASHCLASS
    self.hashclass = hashclass
    if indexclass is None:
      indexclass = choose_indexclass(self.indexbase)
    self.indexclass = indexclass
    self._filemap = None
    self._unindexed = None
    self.index = None
    self._cache = None
    self._indexQ = None
    self._index_Thread = None
    self._monitor_Thread = None

  def init(self):
    ''' Initialise the data dir if not present.
    '''
    statedirpath = self.statedirpath
    if not isdirpath(statedirpath):
      info("mkdir %r", statedirpath)
      with Pfx("mkdir(%r)", statedirpath):
        os.mkdir(statedirpath)
    # create the data subdir if missing
    datadirpath = joinpath(statedirpath, self.DATA_SUBDIR)
    if not isdirpath(datadirpath):
      info("mkdir %r", datadirpath)
      with Pfx("mkdir(%r)", datadirpath):
        os.mkdir(datadirpath)

  def __str__(self):
    return '%s(%s)' % (self.__class__.__name__, shortpath(self.statedirpath))

  def __repr__(self):
    return ( '%s(statedirpath=%r,hashclass=%s,indexclass=%s)'
             % (self.__class__.__name__,
                self.statedirpath,
                self.hashclass.HASHNAME,
                self.indexclass)
           )

  def startup(self):
    ''' Start up the _FilesDir: take locks, start worker threads etc.
    '''
    self._unindexed = {}
    self._filemap = SqliteFilemap(self, self.statefilepath)
    self.runstate.start()
    # cache of open DataFiles
    self._cache = LRU_Cache(
        maxsize=4,
        on_remove=lambda k, datafile: datafile.close()
    )
    # open dbm index
    self.index = self.indexclass(
        self.indexbasepath, self.hashclass,
        self.index_entry_class.from_bytes,
        lock=self._lock)
    self.index.open()
    # set up indexing thread
    # map individual hashcodes to locations before being persistently stored
    # This lets us add data, stash the location in _unindexed and
    # drop the location onto the _indexQ for persistent storage in
    # the index asynchronously.
    self._indexQ = IterableQueue(64)
    T = self._index_Thread = Thread(
        name="%s-index-thread" % (self,),
        target=self._index_updater)
    T.start()
    T = self._monitor_Thread = Thread(
        name="%s-datafile-monitor" % (self,),
        target=self._monitor_datafiles)
    T.start()

  def shutdown(self):
    ''' Shut down the _FilesDir: cancel the runstate, close the
        queues, join the worker threads.
    '''
    self.runstate.cancel()
    self.flush()
    # shut down the monitor Thread
    mon_thread = self._monitor_Thread
    if mon_thread is not None:
      mon_thread.join()
      self._monitor_Thread = None
    # drain index update queue
    Q = self._indexQ
    if Q is not None:
      Q.close()
      self._indexQ = None
    index_thread = self._index_Thread
    if index_thread is not None:
      index_thread.join()
      self._index_Thread = None
    if self._unindexed:
      error("UNINDEXED BLOCKS: %r", self._unindexed)
    # update state to substrate
    self._cache = None
    self._filemap.close()
    self._filemap = None
    self.index.close()
    self.runstate.stop()

  def localpathto(self, rpath):
    ''' Return the path to `rpath`, which is relative to the statedirpath.
    '''
    return joinpath(self.statedirpath, rpath)

  def datapathto(self, rpath):
    ''' Return the path to `rpath`, which is relative to the datadirpath.
    '''
    return joinpath(self.statedirpath, self.DATA_SUBDIR, rpath)

  def state_localpath(self, hashclass):
    ''' Compute the filename of the statefile for the specified `hashclass`.
    '''
    return self.STATE_FILENAME_FORMAT.format(hashname=hashclass.HASHNAME)

  @property
  def statefilepath(self):
    ''' Return the path to the state file.
    '''
    return self.localpathto(self.state_localpath(self.hashclass))

  def get_Archive(self, name=None):
    ''' Return the Archive named `name`.

        If `name` is omitted or `None`
        the Archive path is the `statedirpath`
        plus the extension `'.vt'`.
        Otherwise it is the `statedirpath` plus a dash plus the `name`
        plus the extension `'.vt'`.
        The `name` may not be empty or contain a dot.
    '''
    with Pfx("%s.get_Archive", self):
      if name is None:
        archivepath = self.statedirpath + '.vt'
      else:
        if not name or '.' in name:
          raise ValueError("invalid name: %r" % (name,))
        archivepath = self.statedirpath + '-' + name + '.vt'
      return Archive(archivepath)

  @property
  def indexbase(self):
    ''' Basename of the index.
    '''
    return self.INDEX_FILENAME_BASE_FORMAT.format(hashname=self.hashclass.HASHNAME)

  @property
  def indexbasepath(self):
    ''' Pathname of the index.
    '''
    return self.localpathto(self.indexbase)

  def _queue_index(self, hashcode, entry, post_offset):
    if not isinstance(entry, self.index_entry_class):
      raise RuntimeError("expected %s but got %s %r" % (self.index_entry_class, type(entry), entry))
    with self._lock:
      self._unindexed[hashcode] = entry
    self._indexQ.put( (hashcode, entry, post_offset) )

  def _queue_index_flush(self):
    self._indexQ.put(None)

  @logexc
  def _index_updater(self):
    ''' Thread body to collect hashcode index data from .indexQ and store it.
    '''
    with Pfx("%s._index_updater", self):
      index = self.index
      entry_class = self.index_entry_class
      unindexed = self._unindexed
      filemap = self._filemap
      old_DFstate = None
      indexQ = self._indexQ
      for item in indexQ:
        # dummy item to sync state
        if item is None:
          if old_DFstate is not None:
            filemap.set_indexed_to(old_DFstate.filenum, old_DFstate.indexed_to)
            old_DFstate = None
          continue
        hashcode, entry, post_offset = item
        if not isinstance(entry, entry_class):
          raise RuntimeError("expected %s but got %s %r" % (entry_class, type(entry), entry))
        with self._lock:
          index[hashcode] = entry
          try:
            del unindexed[hashcode]
          except KeyError:
            # this can happen when the same key is indexed twice
            # entirely plausible if a new datafile is added to the datadir
            pass
        DFstate = filemap[entry.n]
        if DFstate is not old_DFstate:
          if old_DFstate is not None:
            filemap.set_indexed_to(old_DFstate.filenum, old_DFstate.indexed_to)
          old_DFstate = DFstate
        DFstate.indexed_to = post_offset
      if old_DFstate is not None:
        filemap.set_indexed_to(old_DFstate.filenum, old_DFstate.indexed_to)

  @locked
  def flush(self):
    ''' Flush all the components.
    '''
    self._queue_index_flush()
    self._cache.flush()
    self.index.flush()

  def __setitem__(self, hashcode, data):
    h = self.add(data)
    if hashcode != h:
      raise ValueError('hashcode %s does not match data, data added under %s instead'
                       % (hashcode, h))

  def __len__(self):
    return len(self.index)

  def hashcodes_from(self, start_hashcode=None, reverse=False):
    ''' Generator yielding the hashcodes from the database in order
        starting with optional `start_hashcode`.

        Parameters:
        * `start_hashcode`: the first hashcode; if missing or None,
          iteration starts with the first key in the index
        * `reverse`: iterate backwards if true, otherwise forwards.
    '''
    unindexed = set(self._unindexed)
    indexed = self.index.hashcodes_from(start_hashcode=start_hashcode,
                                        reverse=reverse)
    unseen_indexed = ( h for h in indexed if h not in unindexed )
    return imerge(sorted(unindexed, reverse=reverse), unseen_indexed)

  def __iter__(self):
    return self.hashcodes_from()

  # without this "in" tries to iterate over the mapping with int indices
  def __contains__(self, hashcode):
    return hashcode in self._unindexed or hashcode in self.index

  def __getitem__(self, hashcode):
    ''' Return the decompressed data associated with the supplied `hashcode`.
    '''
    if not isinstance(hashcode, self.hashclass):
      raise ValueError("hashcode %r is not a %s" % (hashcode, self.hashclass))
    unindexed = self._unindexed
    try:
      entry = unindexed[hashcode]
    except KeyError:
      index = self.index
      try:
        with self._lock:
          entry = index[hashcode]
      except KeyError:
        raise KeyError("%s[%s]: hash not in index" % (self, hashcode))
    try:
      return self.fetch(entry)
    except Exception as e:
      exception("%s[%s]:%s not available: %s", self, hashcode, entry, e)
      raise KeyError(str(hashcode))

  # TODO: memoised BlockMap on demand function?
  def get_blockmap(self, B):
    ''' Return a persistent BlockMap for the supplied Block.
    '''
    raise RuntimeError(
        "%s.get_blockmap: return singleton persistent BlockMap here for Block %s"
        % (self, B))

class SqliteFilemap:
  ''' The file mapping of `n` to `DataFileState`.

      The implementation is an in-memory dict with an SQLite database
      as backing store. SQLite databases are portable across
      architectures and may have multiple users, so that DataDirs
      may be shared.
  '''

  def __init__(self, datadir, path):
    self._lock = Lock()
    self.datadir = datadir
    self.path = path
    self.conn = sqlite3.connect(path, check_same_thread=False)
    self.settings = {}
    self.n_to_DFstate = {}
    self.path_to_DFstate = {}
    c = self.conn.cursor()
    c.execute(r'''
        CREATE TABLE IF NOT EXISTS settings (
            `setting` TEXT,
            `value`   TEXT,
            CONSTRAINT `setting` UNIQUE (`setting`)
        );''')
    c.execute(r'''
        CREATE TABLE IF NOT EXISTS filemap (
            `id`   INTEGER PRIMARY KEY,
            `path` TEXT,
            `indexed_to` INTEGER,
            CONSTRAINT `path` UNIQUE (`path`)
        );''')
    c.connection.commit()
    c.close()
    self._load_map()

  def close(self):
    ''' Close the database.
    '''
    self.conn.close()
    del self.conn

  def _execute(self, sql, *a):
    return self.conn.execute(sql, *a)

  def filenums(self):
    ''' Return the active DFstate filenums.
    '''
    return self.n_to_DFstate.keys()

  def items(self):
    ''' Return the active (n, DFstate) pairs.
    '''
    return self.n_to_DFstate.items()

  def _map(self, path, n, indexed_to=0):
    ''' Add a DataFileState for `path` and `n` to the mapping.
    '''
    datadir = self.datadir
    if n in self.n_to_DFstate:
      warning("replacing n_to_DFstate[%s]", n)
    if path in self.path_to_DFstate:
      warning("replacing path_toDFstate[%r]", path)
    DFstate = DataFileState(
        datadir, n, path, indexed_to=indexed_to)
    self.n_to_DFstate[n] = DFstate
    self.path_to_DFstate[path] = DFstate

  def _load_map(self):
    with self._lock:
      c = self._execute(r'''
          SELECT id, path, indexed_to FROM filemap
      ''')
      for n, path, indexed_to in c.fetchall():
        self._map(path, n, indexed_to)
      c.close()

  def add_path(self, new_path, indexed_to=0):
    ''' Insert a new path into the map.
        Return its DataFileState.
    '''
    conn = self.conn
    with self._lock:
      c = self._execute(r'''
          INSERT INTO filemap(`path`, `indexed_to`) VALUES (?, ?)
      ''', (new_path, 0))
      n = c.lastrowid
      conn.commit()
      self._map(new_path, n, indexed_to=indexed_to)
      c.close()
    return self.n_to_DFstate[n]

  def del_path(self, old_path):
    ''' Forget the information for `old_path`.

        In order to prevent reuse of an id we just set the record's
        `path` and `indexed_to` to NULL.
    '''
    DFstate = self.path_to_DFstate[old_path]
    conn = self.conn
    with self._lock:
      c = self._execute(r'''
          UPDATE filemap SET path=NULL, indexed_to=NULL where id = ?
      ''', (DFstate.filenum,))
      conn.commit()
      c.close()
    del self.n_to_DFstate[DFstate.filenum]
    del self.path_to_DFstate[old_path]

  def __getitem__(self, k):
    if isinstance(k, int):
      return self.n_to_DFstate[k]
    return self.path_to_DFstate[k]

  def get(self, k):
    ''' Return the entry with key `k` or `None`.
    '''
    try:
      return self[k]
    except KeyError:
      return None

  def __contains__(self, k):
    try:
      self[k]
    except KeyError:
      return False
    return True

  def set_indexed_to(self, n, new_indexed_to):
    ''' Update the `indexed_to` value for path `n`.
    '''
    DFstate = self.n_to_DFstate[n]
    conn = self.conn
    with self._lock:
      c = self._execute(r'''
          UPDATE filemap SET indexed_to = ? WHERE id = ?
      ''', (new_indexed_to, n))
      conn.commit()
      c.close()
    DFstate.indexed_to = new_indexed_to

class DataDirIndexEntry(namedtuple('DataDirIndexEntry', 'n offset')):
  ''' A block record for a DataDir.
  '''

  @classmethod
  def from_bytes(cls, data: bytes):
    ''' Parse a binary index entry, return (n, offset).
    '''
    n, offset = get_bs(data)
    file_offset, offset = get_bs(data, offset)
    if offset != len(data):
      raise ValueError(
          "unparsed data from index entry; full entry = %s; n=%d, file_offset=%d, unparsed=%r"
          % (hexlify(data), n, file_offset, data[offset:]))
    return cls(n, file_offset)

  def encode(self) -> bytes:
    ''' Encode (n, offset) to binary form for use as an index entry.
    '''
    return put_bs(self.n) + put_bs(self.offset)

class DataDir(_FilesDir):
  ''' Maintenance of a collection of DataFiles in a directory.

      A DataDir may be used as the Mapping for a MappingStore.

      NB: _not_ thread safe; callers must arrange that.

      The directory may be maintained by multiple instances of this
      class as they will not try to add data to the same DataFile.
      This is intended to address shared Stores such as a Store on
      a NAS presented via NFS, or a Store replicated by an external
      file-level service such as Dropbox or plain old rsync.
  '''

  index_entry_class = DataDirIndexEntry

  def __init__(
      self,
      statedirpath, hashclass,
      *,
      rollover=None,
      **kw
  ):
    ''' Initialise the DataDir with `statedirpath`.

        Parameters:
        * `statedirpath`: a directory containing state information about the
          DataFiles; this is the index-state.csv file and the associated
          index dbm-ish files.
        * `hashclass`: the hash class used to index chunk contents.
        * `indexclass`: the IndexClass providing the index to chunks in the
          DataFiles. If not specified, a supported index class with an
          existing index file will be chosen, otherwise the most favoured
          indexclass available will be chosen.
        * `rollover`: data file roll over size; if a data file grows beyond
          this a new datafile is commenced for new blocks.  Default:
          `DEFAULT_ROLLOVER`.
    '''
    super().__init__(statedirpath, hashclass, **kw)
    if rollover is None:
      rollover = DEFAULT_ROLLOVER
    elif rollover < 1024:
      raise ValueError(
          "rollover < 1024"
          " (a more normal size would be in megabytes or gigabytes): %r"
          % (rollover,))
    self.rollover = rollover
    self._write_n = None
    self._write_DF = None
    self._write_lockpath = None

  def shutdown(self):
    self._close_write_datafile()
    super().shutdown()

  def _open_datafile(self, n):
    ''' Return the DataFileReader with index `n`.
    '''
    cache = self._cache
    DF = cache.get(n)
    if DF is None:
      with self._lock:
        # first, look again now that we have the _lock
        DF = cache.get(n)
        if DF is None:
          # still not in the cache, open the DataFile and put into the cache
          DFstate = self._filemap[n]
          DF = cache[n] = DataFileReader(self.datapathto(DFstate.filename))
          DF.open()
    return DF

  def fetch(self, entry):
    ''' Return the data chunk stored in DataFile `n` at `offset`.
    '''
    DF = self._open_datafile(entry.n)
    return DF.fetch(entry.offset)

  def _get_write_datafile(self):
    ''' Obtain a fresh writable datafile, closing the current one if open.
        Returns (n, WDF).
    '''
    self._close_write_datafile()
    rollover = self.rollover
    WDF = None
    for n, DFstate in self._filemap.items():
      if DFstate.indexed_to < rollover:
        try:
          lockpath = makelockfile(DFstate.pathname, timeout=0)
        except TimeoutError:
          # lock taken, proceed to another file
          continue
      WDF = DataFileWriter(DFstate.pathname)
      break
    if WDF is None:
      # no suitable existing file, make a new one
      while True:
        filename = str(uuid4()) + DATAFILE_DOT_EXT
        pathname = self.datapathto(filename)
        if os.path.exists(pathname):
          error("new datafile path already exists, retrying: %r", pathname)
          continue
        break
      lockpath = makelockfile(pathname, timeout=0)
      WDF = DataFileWriter(pathname, do_create=True)
      DFstate = self._filemap.add_path(filename)
      n = DFstate.filenum
    WDF.open()
    self._write_n = n
    self._write_DF = WDF
    self._write_lockpath = lockpath
    return n, WDF

  def _close_write_datafile(self):
    ''' Close the current writable datafile if open.
    '''
    WDF = self._write_DF
    if WDF is not None:
      WDF.close()
      try:
        os.remove(self._write_lockpath)
      except OSError as e:
        if e.errno == errno.ENOENT:
          warning("remove(%r): %s", self._write_lockpath, e)
        else:
          error("remove(%r): %s", self._write_lockpath, e)
      self._write_n = None
      self._write_DF = None
      self._write_lockpath = None

  def add(self, data):
    ''' Add the supplied data chunk to the current save DataFile,
        return the hashcode.
        Roll the internal state over to a new file if the current
        datafile has reached the rollover threshold.
    '''
    # obtain the write DataFile
    with self._lock:
      WDF = self._write_DF
      if WDF is None:
        n, WDF = self._get_write_datafile()
      else:
        n = self._write_n
    # save the data chunk
    with WDF:
      offset, post_offset = WDF.add(data)
    # queue the index update
    hashcode = self.hashclass.from_chunk(data)
    self._queue_index(hashcode, DataDirIndexEntry(n, offset), post_offset)
    # see if the write DataFile is now full; if so, close it
    rollover = self.rollover
    with self._lock:
      if rollover is not None and post_offset >= rollover:
        self._close_write_datafile()
    return hashcode

  @staticmethod
  def scanfrom(filepath, offset=0, **kw):
    ''' Scan the specified `filepath` from `offset`, yielding data chunks.
    '''
    with DataFileReader(filepath) as DF:
      yield from DF.scanfrom(offset, **kw)

  def _monitor_datafiles(self):
    ''' Thread body to poll all the datafiles regularly for new data arrival.

        This is what supports shared use of the data area. Other clients
        may write to their onw datafiles and this thread sees new files
        and new data in existing files and scans them, adding the index
        information to the local state.
    '''
    filemap = self._filemap
    indexQ = self._indexQ
    datadirpath = joinpath(self.statedirpath, self.DATA_SUBDIR)
    while not self.cancelled:
      if self.flag_scan_disable:
        time.sleep(1)
        continue
      # scan for new datafiles
      with Pfx("listdir(%r)", datadirpath):
        try:
          listing = list(os.listdir(datadirpath))
        except OSError as e:
          if e.errno == errno.ENOENT:
            error("listing failed: %s", e)
            time.sleep(2)
            continue
          raise
        for filename in listing:
          if (
              not filename.startswith('.')
              and filename.endswith(DATAFILE_DOT_EXT)
              and filename not in filemap
          ):
            info("MONITOR: add new filename %r", filename)
            filemap.add_path(filename)
      # now scan known datafiles for new data
      for filenum in self._filemap.filenums():
        if self.cancelled or self.flag_scan_disable:
          break
        # don't monitor the current datafile: our own actions will update it
        n = self._write_n
        if n is not None and filenum == n:
          continue
        try:
          DFstate = filemap[filenum]
        except KeyError:
          warning("missing entry %d in filemap", filenum)
          continue
        with Pfx(DFstate.filename):
          try:
            new_size = DFstate.stat_size()
          except OSError as e:
            warning("stat: %s", e)
            continue
          else:
            if new_size is None:
              info("skip nonfile")
              continue
          if new_size > DFstate.scanned_to:
            offset = DFstate.scanned_to
            for record, post_offset in DFstate.scanfrom(offset=offset):
              hashcode = self.hashclass.from_chunk(record.data)
              indexQ.put( (hashcode, DataDirIndexEntry(filenum, offset), post_offset) )
              DFstate.scanned_to = post_offset
              if self.cancelled:
                break
              offset = post_offset
            self.flush()
        self.flush()
      time.sleep(1)

class PlatonicDirIndexEntry(namedtuple('PlatonicDirIndexEntry', 'n offset length')):
  ''' A block record for a PlatonicDir.
  '''

  @classmethod
  def from_bytes(cls, data: bytes):
    ''' Parse a binary index entry, return (n, offset).
    '''
    n, offset = get_bs(data)
    file_offset, offset = get_bs(data, offset)
    length, offset = get_bs(data, offset)
    if offset != len(data):
      raise ValueError("unparsed data from index entry; full entry = %s" % (hexlify(data),))
    return cls(n, file_offset, length)

  def encode(self) -> bytes:
    ''' Encode (n, offset) to binary form for use as an index entry.
    '''
    return put_bs(self.n) + put_bs(self.offset) + put_bs(self.length)

class PlatonicFile(MultiOpenMixin, ReadMixin):
  ''' A PlatonicFile is a normal file whose content is used as the
      reference for block data.
  '''

  def __init__(self, path):
    MultiOpenMixin.__init__(self)
    self.path = path
    self._fd = None
    # dummy value since all I/O goes through datafrom, which uses pread
    self._seek_offset = 0

  def __str__(self):
    return "PlatonicFile(%s)" % (shortpath(self.path,))

  def startup(self):
    ''' Startup: open the file for read.
    '''
    self._fd = os.open(self.path, os.O_RDONLY)

  def shutdown(self):
    ''' Shutdown: close the file.
    '''
    os.close(self._fd)
    self._fd = None

  def tell(self):
    ''' Return the notional file offset.
    '''
    return self._seek_offset

  def seek(self, pos, how=SEEK_SET):
    ''' Adjust the notional file offset.
    '''
    if how == SEEK_SET:
      pass
    elif how == SEEK_CUR:
      pos += self._seek_offset
    elif how == SEEK_END:
      pos += os.fstat(self._fd).st_size
    else:
      raise ValueError("unsupported seek how value: 0x%02x" % (how,))
    if pos < 0:
      raise ValueError("seek out of range")
    self._seek_offset = pos

  def datafrom(self, offset, readsize=None):
    ''' Return an iterable of data from this file from `offset`.
    '''
    if readsize is None:
      readsize = DEFAULT_READSIZE
    return datafrom_fd(self._fd, offset, readsize)

  def fetch(self, offset, length):
    ''' Fetch the data from this file at `offset`.
    '''
    data = self.read(length, offset=offset, longread=True)
    if len(data) != length:
      raise RuntimeError(
          "%r: asked for %d bytes from offset %d, but got %d"
          % (self.path, length, offset, len(data)))
    return data

class PlatonicDir(_FilesDir):
  ''' Presentation of a block map based on a raw directory tree of
      files such as a preexisting media server.

      A PlatonicDir may be used as the Mapping for a MappingStore.

      NB: _not_ thread safe; callers must arrange that.

      A PlatonicDir is read-only. Data blocks are fetched directly
      from the files in the backing directory tree.
  '''

  # delays during scanning to limit the CPU and I/O impact of the
  # monitoring and updating
  DELAY_INTERSCAN = 1.0     # regular pause between directory scans
  DELAY_INTRASCAN = 0.1     # stalls during scan: per directory and after big files

  index_entry_class = PlatonicDirIndexEntry

  def __init__(
      self,
      statedirpath, hashclass,
      *,
      exclude_dir=None, exclude_file=None,
      follow_symlinks=False, archive=None, meta_store=None,
      **kw
  ):
    ''' Initialise the PlatonicDir with `statedirpath` and `datadirpath`.

        Parameters:
        * `statedirpath`: a directory containing state information about the
          DataFiles; this is the index-state.sqlite file and the associated
          index dbm-ish files.
        * `hashclass`: the hash class used to index chunk contents.
        * `exclude_dir`: optional function to test a directory path for
          exclusion from monitoring; default is to exclude directories
          whose basename commences with a dot.
        * `exclude_file`: optional function to test a file path for
          exclusion from monitoring; default is to exclude directories
          whose basename commences with a dot.
        * `follow_symlinks`: follow symbolic links, default False.
        * `meta_store`: an optional Store used to maintain a Dir
          representing the ideal directory; unhashed data blocks
          encountered during scans which are promoted to HashCodeBlocks
          are also stored here
        * `archive`: optional Archive ducktype instance with a
          .update(Dirent[,when]) method

        Other keyword arguments are passed to _FilesDir.__init__.

        The directory and file paths tested are relative to the
        data directory path.
    '''
    if meta_store is None:
      raise ValueError("meta_store may not be None")
    super().__init__(statedirpath, hashclass, **kw)
    if exclude_dir is None:
      exclude_dir = self._default_exclude_path
    if exclude_file is None:
      exclude_file = self._default_exclude_path
    self.exclude_dir = exclude_dir
    self.exclude_file = exclude_file
    self.follow_symlinks = follow_symlinks
    self.meta_store = meta_store
    if meta_store is not None and archive is None:
      archive = super().get_Archive()
    elif archive is not None:
      if isinstance(archive, str):
        archive = Archive(archive)
    self.archive = archive
    self.topdir = None

  def startup(self):
    if self.meta_store is not None:
      self.meta_store.open()
      archive = self.archive
      _, D = archive.last
      if D is None:
        info("%r: no archive entries, create empty topdir Dir", archive)
        D = Dir('.')
        archive.update(D)
      self.topdir = D
    super().startup()

  def shutdown(self):
    if self.meta_store is not None:
      self.sync_meta()
      self.meta_store.close()
    super().shutdown()

  def sync_meta(self):
    ''' Update the Archive state.
    '''
    # update the topdir state before any save
    if self.meta_store is not None:
      with self.meta_store:
        self.archive.update(self.topdir)
        ##dump_Dirent(self.topdir, recurse=True)

  @staticmethod
  def _default_exclude_path(path):
    ''' Default function to exclude a path from the file tree traversal.
    '''
    base = basename(path)
    return not base or base.startswith('.')

  def _open_datafile(self, n):
    ''' Return the DataFile with index `n`.
    '''
    cache = self._cache
    DF = cache.get(n)
    if DF is None:
      with self._lock:
        # first, look again now that we have the _lock
        DF = cache.get(n)
        if DF is None:
          # still not in the cache, open the DataFile and put into the cache
          DFstate = self._filemap[n]
          DF = cache[n] = PlatonicFile(self.datapathto(DFstate.filename))
          DF.open()
    return DF

  def fetch(self, entry):
    ''' Return the data chunk stored in DataFile `n` at `offset`.
    '''
    DF = self._open_datafile(entry.n)
    return DF.fetch(entry.offset, entry.length)

  @logexc
  def _monitor_datafiles(self):
    ''' Thread body to poll the ideal tree for new or changed files.
    '''
    meta_store = self.meta_store
    filemap = self._filemap
    indexQ = self._indexQ
    datadirpath = joinpath(self.statedirpath, self.DATA_SUBDIR)
    if meta_store is not None:
      topdir = self.topdir
    else:
      warning("%s: no meta_store!", self)
    while not self.cancelled:
      time.sleep(self.DELAY_INTERSCAN)
      if self.flag_scan_disable:
        continue
      # scan for new datafiles
      with Pfx("%r", datadirpath):
        seen = set()
        info("scan tree...")
        for dirpath, dirnames, filenames in os.walk(datadirpath, followlinks=True):
          time.sleep(self.DELAY_INTRASCAN)
          if self.cancelled or self.flag_scan_disable:
            break
          rdirpath = relpath(dirpath, datadirpath)
          with Pfx(rdirpath):
            # this will be the subdirectories into which to recurse
            pruned_dirnames = []
            for dname in dirnames:
              if self.exclude_dir(joinpath(rdirpath, dname)):
                # unwanted
                continue
              subdirpath = joinpath(dirpath, dname)
              try:
                S = os.stat(subdirpath)
              except OSError as e:
                # inaccessable
                warning("stat(%r): %s, skipping", subdirpath, e)
                continue
              ino = S.st_dev, S.st_ino
              if ino in seen:
                # we have seen this subdir before, probably via a symlink
                # TODO: preserve symlinks? attach alter ego directly as a Dir?
                debug("seen %r (dev=%s,ino=%s), skipping", subdirpath, ino[0], ino[1])
                continue
              seen.add(ino)
              pruned_dirnames.append(dname)
            dirnames[:] = pruned_dirnames
            if meta_store is None:
              warning("no meta_store")
              D = None
            else:
              with meta_store:
                D = topdir.makedirs(rdirpath, force=True)
                # prune removed names
                names = list(D.keys())
                for name in names:
                  if name not in dirnames and name not in filenames:
                    info("del %r", name)
                    del D[name]
            for filename in filenames:
              with Pfx(filename):
                if self.cancelled or self.flag_scan_disable:
                  break
                rfilepath = joinpath(rdirpath, filename)
                if self.exclude_file(rfilepath):
                  continue
                filepath = joinpath(dirpath, filename)
                if not isfilepath(filepath):
                  continue
                # look up this file in our file state index
                DFstate = filemap.get(rfilepath)
                if (
                    DFstate is not None
                    and D is not None
                    and filename not in D
                ):
                  # in filemap, but not in dir: start again
                  warning("in filemap but not in Dir, rescanning")
                  filemap.del_path(rfilepath)
                  DFstate = None
                if DFstate is None:
                  XP("new DFstate")
                  DFstate = filemap.add_path(rfilepath)
                try:
                  new_size = DFstate.stat_size(self.follow_symlinks)
                except OSError as e:
                  if e.errno == errno.ENOENT:
                    warning("forgetting missing file")
                    self._del_datafilestate(DFstate)
                  else:
                    warning("stat: %s", e)
                  continue
                if new_size is None:
                  # skip non files
                  debug("SKIP non-file")
                  continue
                if meta_store:
                  try:
                    E = D[filename]
                  except KeyError:
                    E = FileDirent(filename)
                    D[filename] = E
                  else:
                    if not E.isfile:
                      info("new FileDirent replacing previous nonfile: %s", E)
                      E = FileDirent(filename)
                      D[filename] = E
                if new_size > DFstate.scanned_to:
                  if DFstate.scanned_to > 0:
                    info("scan from %d", DFstate.scanned_to)
                  if meta_store is not None:
                    blockQ = IterableQueue()
                    R = meta_store._defer(
                        lambda B, Q: top_block_for(spliced_blocks(B, Q)),
                        E.block, blockQ)
                  scan_from = DFstate.scanned_to
                  scan_start = time.time()
                  for offset, flags, data, post_offset in DFstate.scanfrom(
                      offset=DFstate.scanned_to, do_decompress=True,
                  ):
                    assert flags == 0
                    hashcode = self.hashclass.from_chunk(data)
                    indexQ.put( (
                        hashcode,
                        PlatonicDirIndexEntry(DFstate.filenum, offset, len(data)),
                        post_offset
                    ) )
                    if meta_store is not None:
                      B = Block(data=data, hashcode=hashcode, added=True)
                      blockQ.put( (offset, B) )
                    DFstate.scanned_to = post_offset
                    if self.cancelled or self.flag_scan_disable:
                      break
                  elapsed = time.time() - scan_start
                  scanned = DFstate.scanned_to - scan_from
                  if elapsed > 0:
                    scan_rate = scanned / elapsed
                  else:
                    scan_rate = None
                  if scan_rate is None:
                    info(
                        "scanned to %d: %s",
                        DFstate.scanned_to,
                        transcribe_bytes_geek(scanned))
                  else:
                    info(
                        "scanned to %d: %s at %s/s",
                        DFstate.scanned_to,
                        transcribe_bytes_geek(scanned),
                        transcribe_bytes_geek(scan_rate))
                  if meta_store is not None:
                    blockQ.close()
                    try:
                      top_block = R()
                    except MissingHashcodeError as e:
                      error("missing data, forcing rescan: %s", e)
                      DFstate.scanned_to = 0
                    else:
                      E.block = top_block
                      D.changed = True
                      self.sync_meta()
                  # stall after a file scan, briefly, to limit impact
                  if elapsed > 0:
                    time.sleep(min(elapsed, self.DELAY_INTRASCAN))
      self.flush()

  @staticmethod
  def scanfrom(filepath, offset=0, do_decompress=False):
    ''' Scan the specified `filepath` from `offset`, yielding data chunks.
    '''
    assert do_decompress is True
    scanner = scanner_from_filename(filepath)
    with open(filepath, 'rb') as fp:
      fp.seek(offset)
      for data in blocked_chunks_of(read_from(fp, DEFAULT_SCAN_SIZE), scanner):
        post_offset = offset + len(data)
        yield offset, 0, data, post_offset
        offset = post_offset

if __name__ == '__main__':
  from .datadir_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Block stores.
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Basic Store classes.
'''

from __future__ import with_statement
from abc import ABC, abstractmethod
from functools import partial
from os.path import expanduser, isabs as isabspath
import sys
from threading import Semaphore
from cs.later import Later, SubLater
from cs.logutils import warning, error
from cs.pfx import Pfx
from cs.progress import Progress
from cs.py.func import prop, funcname
from cs.queues import Channel, IterableQueue
from cs.resources import MultiOpenMixin, RunStateMixin, RunState
from cs.result import report
from cs.seq import Seq
from cs.threads import bg
from cs.x import X
from . import defaults, Lock
from .datadir import DataDir, PlatonicDir
from .hash import DEFAULT_HASHCLASS, HashCodeUtilsMixin, MissingHashcodeError

class StoreError(Exception):
  ''' Raised by Store operation failures.
  '''
  pass

class _BasicStoreCommon(MultiOpenMixin, HashCodeUtilsMixin, RunStateMixin, ABC):
  ''' Core functions provided by all Stores.

      Subclasses should not subclass this class but BasicStoreSync
      or BasicStoreAsync; these provide the *_bg or non-*_bg sibling
      methods of those described below so that a subclass need only
      implement the synchronous or asynchronous forms. Most local
      Stores will derive from BasicStoreSync and remote Stores
      derive from BasicStoreAsync.

      A subclass should provide thread-safe implementations of the following
      methods:

        .add(block) -> hashcode
        .get(hashcode, [default=None]) -> block (or default)
        .contains(hashcode) -> boolean
        .flush()

      A subclass _may_ provide thread-safe implementations of the following
      methods:

        .hashcodes(starting_hashcode, length) -> iterable-of-hashcodes

      The background (*_bg) functions return cs.later.LateFunction instances
      for deferred collection of the operation result.

      A convenience .lock attribute is provided for simple mutex use.

      The .readonly attribute may be set to prevent writes and trap
      surprises; it relies on assert statements.

      The .writeonly attribute may be set to trap surprises when no blocks
      are expected to be fetched; it relies on asssert statements.

      The mapping special methods __getitem__ and __contains__ call
      the implementation methods .get() and .contains().
  '''

  _seq = Seq()

  def __init__(self, name, capacity=None, hashclass=None, lock=None, runstate=None):
    ''' Initialise the Store.

        Parameters:
        * `name`: a name for this Store;
          if None, a sequential name based on the Store class name
          is generated
        * `capacity`: a capacity for the internal Later queue, default 4
        * `hashclass`: the hash class to use for this Store,
          default: `DEFAULT_HASHCLASS`
        * `lock`: an optional lock for managing concurrency,
          if not supplied a new `threading.RLock` is allocated
        * `runstate`: a `cs.resources.RunState` for external control;
          if not supplied one is allocated
    '''
    with Pfx("_BasicStoreCommon.__init__(%s,..)", name):
      if not isinstance(name, str):
        raise TypeError(
            "initial `name` argument must be a str, got %s"
            % (type(name),))
      if name is None:
        name = "%s%d" % (type(self).__name__, next(_BasicStoreCommon._seq()))
      if capacity is None:
        capacity = 4
      if hashclass is None:
        hashclass = DEFAULT_HASHCLASS
      if runstate is None:
        runstate = RunState(name)
      MultiOpenMixin.__init__(self, lock=lock)
      RunStateMixin.__init__(self, runstate=runstate)
      self._attrs = {}
      self.name = name
      self._capacity = capacity
      self.hashclass = hashclass
      self.config = None
      self.logfp = None
      self.mountdir = None
      self.readonly = False
      self.writeonly = False
      self._archives = {}
      self._blockmapdir = None
      self.block_cache = None

  def __str__(self):
    ##return "STORE(%s:%s)" % (type(self), self.name)
    params = []
    for attr, val in sorted(self._attrs.items()):
      params.append(attr + '=' + str(val))
    return "%s:%s(%s)" % (
        self.__class__.__name__, self.hashclass.HASHNAME,
        ','.join([repr(self.name)] + params)
    )

  __repr__ = __str__

  __bool__ = lambda self: True

  # Basic support for putting Stores in sets.
  def __hash__(self):
    return id(self)

  def hash(self, data):
    ''' Return a Hash object from data bytes.
        NB: this does _not_ store the data.
    '''
    return self.hashclass.from_chunk(data)

  # Stores are equal only to themselves.
  def __eq__(self, other):
    return self is other

  ###################
  ## Mapping methods.
  ##

  def __contains__(self, h):
    ''' Test if the supplied hashcode is present in the store.
    '''
    return self.contains(h)

  def __iter__(self):
    return self.hashcodes_from()

  def keys(self):
    ''' Return an iterator over the Store's hashcodes.
    '''
    return iter(self)

  def __getitem__(self, h):
    ''' Return the data bytes associated with the supplied hashcode.
        Raise KeyError if the hashcode is not present.
    '''
    block = self.get(h)
    if block is None:
      raise MissingHashcodeError(h)
    return block

  def __setitem__(self, h, data):
    ''' Save `data` against hash key `h`.
        Actually saves the data against the Store's hash function
        and raises ValueError if that does not match the supplied
        `h`.
    '''
    h2 = self.add(data)
    if h != h2:
      raise ValueError("h:%s != hash(data):%s" % (h, h2))

  ###########################
  ## Context manager methods.
  ##

  def __enter__(self):
    defaults.pushStore(self)
    return MultiOpenMixin.__enter__(self)

  def __exit__(self, exc_type, exc_value, traceback):
    if exc_value:
      import traceback as TB
      TB.print_tb(traceback, file=sys.stderr)
    defaults.popStore()
    return MultiOpenMixin.__exit__(self, exc_type, exc_value, traceback)

  ##########################
  ## MultiOpenMixin methods.
  ##

  def startup(self):
    ''' Start the Store.
    '''
    self.runstate.start()
    self.__funcQ = Later(self._capacity, name="%s:Later(__funcQ)" % (self.name,))
    self._worker = SubLater(self.__funcQ)
    self._reaper = self._worker.reaper()

  def shutdown(self):
    ''' Called by final MultiOpenMixin.close().
    '''
    self.runstate.cancel()
    self._worker.close()
    self._reaper.join()
    L = self.__funcQ
    L.shutdown()
    L.wait()
    del self.__funcQ
    self.runstate.stop()

  #############################
  ## Function dispatch methods.
  ##

  def _defer(self, func, *args, **kwargs):
    ''' Defer a function via the internal Later queue.
    '''
    self.open()
    def deferred():
      with self:
        result = func(*args, **kwargs)
      return result
    deferred.__name__ = "deferred:" + funcname(func)
    LF = self._worker.defer(deferred)
    LF.notify(lambda LF: self.close())
    return LF

  ##########################################################################
  # Core Store methods, all abstract.
  @abstractmethod
  def add(self, data):
    ''' Add the `data` to the Store, return its hashcode.
    '''
    raise NotImplementedError()

  @abstractmethod
  def add_bg(self, data):
    ''' Dispatch the add request in the backgrounmd, return Result.
    '''
    raise NotImplementedError()

  @abstractmethod
  def get(self, h):
    ''' Fetch the data for hashcode `h` from the Store, or None.
    '''
    raise NotImplementedError()

  @abstractmethod
  def get_bg(self, h):
    ''' Dispatch the get request in the backgrounmd, return Result.
    '''
    raise NotImplementedError()

  @abstractmethod
  def contains(self, h):
    ''' Test whether the hashcode `h` is present in the Store.
    '''
    raise NotImplementedError()

  @abstractmethod
  def contains_bg(self, h):
    ''' Dispatch the contains request in the backgrounmd, return Result.
    '''
    raise NotImplementedError()

  @abstractmethod
  def flush(self):
    ''' Flush outstanding tasks to the next lowest abstraction.
    '''
    raise NotImplementedError()

  @abstractmethod
  def flush_bg(self):
    ''' Dispatch the flush request in the backgrounmd, return Result.
    '''
    raise NotImplementedError()

  ##########################################################################
  # Archive support.
  def add_archive(self, name, archive):
    ''' Add an `archive` by `name`.
    '''
    archives = self._archives
    with self._lock:
      if name in archives:
        raise KeyError("archive named %r already exists" % (name,))
      archives[name] = archive

  def get_archive(self, name):
    ''' Fetch the named archive or None.
    '''
    return self._archives.get(name)

  ##########################################################################
  # Blockmaps.
  @prop
  def blockmapdir(self):
    ''' The path to this Store's blockmap directory, if specified.
    '''
    with Pfx("%s.blockmapdir", self):
      dirpath = self._blockmapdir
      if dirpath is None:
        cfg = self.config
        dirpath = cfg.get_default('blockmapdir')
        if dirpath is not None:
          if dirpath.startswith('['):
            endpos = dirpath.find(']', 1)
            if endpos < 0:
              warning('[GLOBAL].blockmapdir: starts with "[" but no "]": %r', dirpath)
            else:
              clausename = dirpath[1:endpos].strip()
              with Pfx('[%s]', clausename):
                if not clausename:
                  warning('[GLOBAL].blockmapdir: empty clause name: %r', dirpath)
                else:
                  try:
                    S = cfg[clausename]
                  except KeyError:
                    warning("unknown config clause")
                  else:
                    rdirpathpos = endpos + 1
                    if rdirpathpos == len(dirpath):
                      rdirpath = 'blockmaps'
                    elif dirpath.startswith('/', rdirpathpos):
                      rdirpath = dirpath[rdirpathpos+1:]
                      if not rdirpath:
                        rdirpath = 'blockmaps'
                    else:
                      warning(
                          '[GLOBAL].blockmapdir: %r not followed with a slash: %r',
                          dirpath[:endpos+1], dirpath)
                      rdirpath = None
                    if rdirpath:
                      dirpath = S.localpathto(rdirpath)
          else:
            # TODO: generic handler for Store subpaths needed
            if not isabspath(dirpath):
              dirpath = expanduser(dirpath)
              if not isabspath(dirpath):
                dirpath = S.localpathto(dirpath)
      return dirpath

  @blockmapdir.setter
  def blockmapdir(self, dirpath):
    ''' Set the Blockmap directory path.
    '''
    self._blockmapdir = dirpath

  def pushto(
      self, S2,
      capacity=1024, block_progress=None, bytes_progress=None
  ):
    ''' Allocate a Queue for Blocks to push from this Store to another Store `S2`.
        Return (Q, T) where `Q` is the new Queue and `T` is the
        Thread processing the Queue.

        Parameters:
        * `S2`: the secondary Store to receive Blocks.
        * `capacity`: the Queue capacity, arbitrary default 1024.
        * `block_progress`: an optional Progress counting submitted and completed Blocks.
        * `bytes_progress`: an optional Progress counting submitted and completed data bytes.

        Once called, the caller can then .put Blocks onto the Queue.
        When finished, call Q.close() to indicate end of Blocks and
        T.join() to wait for the processing completion.
    '''
    if capacity < 1:
      raise ValueError("capacity must be >= 1, got: %r" % (capacity,))
    lock = Lock()
    sem = Semaphore(capacity)
    ##sem = Semaphore(1)
    if block_progress is None:
      added_block = lambda B: None
      did_block = lambda B: None
    else:
      def added_block(B):
        ''' Advance the Block total.
        '''
        with lock:
          block_progress.total += 1
      def did_block(B):
        ''' Advance the Block progress counter.
        '''
        with lock:
          block_progress.position += 1
    if bytes_progress is None:
      added_bytes = lambda B: None
      did_bytes = lambda B: None
    else:
      def added_bytes(B):
        ''' Advance the bytes total.
        '''
        with lock:
          bytes_progress.total += B.span
      def did_bytes(B):
        ''' Advance the bytes progress counter.
        '''
        with lock:
          bytes_progress.position += B.span
    def Xs(s):
      ''' Terse inline string debug output.
      '''
      print(s, file=sys.stderr, end='', flush=True)
    X("NEW PUSHTO %r ==> %r", self.name, S2.name)
    name = "%s.pushto(%s)" % (self.name, S2.name)
    with Pfx(name):
      Q = IterableQueue(capacity=capacity, name=name)
      S1 = self
      S1.open()
      S2.open()
      pending = set()
      def worker(name, Q, S1, S2, sem):
        ''' This is the worker function which pushes Blocks from
            the Queue to the second Store.
        '''
        X("START PUSHTO PROCESSING...")
        with Pfx("%s: worker", name):
          with S1:
            for B in Q:
              Xs("B")
              added_block(B)
              added_bytes(B)
              with Pfx("%s", B):
                try:
                  h = B.hashcode
                except AttributeError:
                  warning("not a hashcode Block, skipping")
                  did_block(B)
                  did_bytes(B)
                  continue
                def addblock(S1, S2, h, B):
                  ''' Add the Block `B` to `S2` if not present.
                  '''
                  Xs("?")
                  if S2.contains(h):
                    return False
                  try:
                    data = S1[h]
                  except KeyError as e:
                    error("missing %s[%s]: %s", S1.name, h, e)
                    return None
                  Xs("+")
                  S2.add(data)
                  return True
                Xs("{")
                sem.acquire()
                addR = S2._defer(addblock, S1, S2, h, B)
                Xs("<")
                with lock:
                  pending.add(addR)
                def after_add(addR):
                  ''' Forget that `addR` is pending.
                      This will be called after `addR` completes.
                  '''
                  Xs(">")
                  with lock:
                    pending.remove(addR)
                  did_block(B)
                  did_bytes(B)
                  Xs("}")
                  sem.release()
                addR.notify(after_add)
            X("PUSHTO: NO MORE BLOCKS")
            with lock:
              outstanding = list(pending)
            X("PUSHTO: %d outstanding, waiting...", len(outstanding))
            for R in outstanding:
              R.join()
          S2.close()
          S1.close()
        X("PUSHTO: PROCESSING THREAD COMPLETES")
      T = bg(partial(worker, name, Q, S1, S2, sem))
      return Q, T

class BasicStoreSync(_BasicStoreCommon):
  ''' Subclass of _BasicStoreCommon expecting synchronous operations
      and providing asynchronous hooks, dual of BasicStoreAsync.
  '''

  #####################################
  ## Background versions of operations.
  ##

  def add_bg(self, data):
    return self._defer(self.add, data)

  def get_bg(self, h):
    return self._defer(self.get, h)

  def contains_bg(self, h):
    return self._defer(self.contains, h)

  def flush_bg(self):
    return self._defer(self.flush)

class BasicStoreAsync(_BasicStoreCommon):
  ''' Subclass of _BasicStoreCommon expecting asynchronous operations
      and providing synchronous hooks, dual of BasicStoreSync.
  '''

  #####################################
  ## Background versions of operations.
  ##

  def add(self, data):
    return self.add_bg(data)()

  def get(self, h):
    return self.get_bg(h)()

  def contains(self, h):
    return self.contains_bg(h)()

  def flush(self):
    return self.flush_bg()()

class MappingStore(BasicStoreSync):
  ''' A Store built on an arbitrary mapping object.
  '''

  def __init__(self, name, mapping, **kw):
    BasicStoreSync.__init__(self, name, **kw)
    self.mapping = mapping
    self._attrs.update(mapping=mapping)

  def startup(self):
    super().startup()
    mapping = self.mapping
    try:
      openmap = mapping.open
    except AttributeError:
      pass
    else:
      openmap()

  def shutdown(self):
    mapping = self.mapping
    try:
      closemap = mapping.close
    except AttributeError:
      pass
    else:
      closemap()
    super().shutdown()

  def add(self, data):
    with Pfx("add %d bytes", len(data)):
      mapping = self.mapping
      h = self.hash(data)
      if h not in mapping:
        mapping[h] = data
      else:
        if False:
          with Pfx("EXISTING HASH"):
            try:
              data2 = mapping[h]
            except Exception as e:
              error("fetch FAILED: %s", e)
            else:
              if data != data2:
                warning("data mismatch: .add data=%r, Store data=%r", data, data2)
      return h

  def get(self, h, default=None):
    try:
      data = self.mapping[h]
    except KeyError:
      return default
    return data

  def contains(self, h):
    return h in self.mapping

  def flush(self):
    ''' Call the .flush method of the underlying mapping, if any.
    '''
    map_flush = getattr(self.mapping, 'flush', None)
    if map_flush is not None:
      map_flush()

  def __len__(self):
    return len(self.mapping)

  def __iter__(self):
    ''' Return iterator over the mapping; required for use of HashCodeUtilsMixin.hashcodes_from.
    '''
    return iter(self.mapping)

  def hashcodes_from(self, start_hashcode=None, reverse=False):
    ''' Use the mapping's .hashcodes_from if present, otherwise use
        HashCodeUtilsMixin.hashcodes_from.
    '''
    try:
      hashcodes_method = self.mapping.hashcodes_from
    except AttributeError:
      return HashCodeUtilsMixin.hashcodes_from(self, start_hashcode=start_hashcode, reverse=reverse)
    return hashcodes_method(start_hashcode=start_hashcode, reverse=reverse)

class ProxyStore(BasicStoreSync):
  ''' A Store managing various subsidiary Stores.

      Three classes of Stores are managed:

      Save stores. All data added to the Proxy is added to these Stores.

      Read Stores. Requested data may be obtained from these Stores.

      Copy Stores. Data retrieved from a `read2` Store is copied to these Stores.

      A example setup utilising a working ProxyStore might look like this:

          ProxyStore(
            save=[local,upstream],
            save2=[spool],
            read=[local],
            read2=[upstream],
            copy2=[local],
          )

      In this example:
      * `local`: is a local low latency store such as a DataDirStore.
      * `upstream`: is a remote high latency Store such as a TCPStore.
      * `spool`: is a local scondary Store, probably a DataDirStore.

      This setup causes all saved data to be saved to `local` and
      `upstream`.
      If a save to `local` or `upstream` fails,
      for example if the upstream if offline,
      the save is repeated to the `spool`,
      intended as a holding location for data needing a resave.

      Reads are attempted first from the `read` Stores, then from
      the `read2` Stores.
      If there are any `copy2` Stores,
      any data obtained from `read2` are copied into the `copy2` Stores;
      in this way remote data become locally saved.

      TODO: replay and purge the spool? probably better as a separate
      pushto operation ("vt despool spool_store upstream_store").
  '''

  def __init__(self, name, save, read, *, save2=(), read2=(), copy2=(), **kw):
    ''' Initialise a ProxyStore.

        Parameters:
        * `name`: ProxyStore name.
        * `save`: iterable of Stores to which to save blocks
        * `read`: iterable of Stores from which to fetch blocks
        * `save2`: fallback Store for saves which fail
        * `read2`: optional fallback iterable of Stores from which
          to fetch blocks if not found via `read`. Typically these
          would be higher latency upstream Stores.
        * `copy2`: optional iterable of Stores to receive copies
          of data obtained via `read2` Stores.
    '''
    BasicStoreSync.__init__(self, name, **kw)
    self.save = frozenset(save)
    self.read = frozenset(read)
    self.save2 = frozenset(save2)
    self.read2 = frozenset(read2)
    self.copy2 = frozenset(copy2)
    self._attrs.update(save=save, read=read)
    if save2:
      self._attrs.update(save2=save2)
    if read2:
      self._attrs.update(read2=read2)
    if copy2:
      self._attrs.update(copy2=copy2)
    self.readonly = len(self.save) == 0

  def __str__(self):
    return "%s(%r)" % (type(self).__name__, self.name)

  def startup(self):
    super().startup()
    for S in self.save | self.read | self.save2 | self.read2 | self.copy2:
      S.open()

  def shutdown(self):
    for S in self.save | self.read | self.save2 | self.read2 | self.copy2:
      S.close()
    super().shutdown()

  @staticmethod
  def _multicall0(stores, method_name, args):
    ''' Basic multicall of _bg methods yielding (LF, S) pairs in the order submitted.
    '''
    assert method_name.endswith('_bg')
    stores = list(stores)
    for S in stores:
      with Pfx("%s.%s()", S, method_name):
        with S:
          LF = getattr(S, method_name)(*args)
      yield LF, S   # outside Pfx because this is a generator

  def _multicall(self, stores, method_name, args):
    ''' Generator yielding (S, result, exc_info) for each call to
        S.method_name(args) in the order completed.

        The method_name should be one of the *_bg names which return
        LateFunctions.
        Methods are called in parallel and values returned as
        completed, so the (S, LateFUnction) returns may not be in the same
        order as the supplied `stores`.
        `stores`: iterable of Stores on which to call `method_name`
        `method_name`: name of Store method
        `args`: positional arguments for the method call
    '''
    LFmap = dict(self._multicall0(stores, method_name, args))
    for LF in report(LFmap.keys()):
      # locate the corresponding store for context
      S = LFmap[LF]
      yield S, LF.result, LF.exc_info

  def add(self, data):
    ''' Add a data chunk to the save Stores.
        This queues all the saves in the background and returns the
        hashcode received.
    '''
    ch = Channel()
    self._defer(self._bg_add, data, ch)
    hashcode = ch.get()
    if hashcode is None:
      raise RuntimeError("no hashcode returned from .add")
    return hashcode

  def _bg_add(self, data, ch):
    ''' Add a data chunk to the save Stores.

        Parameters:
        `data`: the data to add
        `ch`: a channel for hashcode return
    '''
    assert ch, "ch is false! %r" % (ch,)
    try:
      if not self.save:
        # no save - allow add if hashcode already present - dubious
        hashcode = self.hash(data)
        if hashcode in self:
          ch.put(hashcode)
          ch = None
          return
        raise RuntimeError("new add but no save Stores")
      ok = True
      fallback = None
      for S, hashcode, exc_info in self._multicall(self.save, 'add_bg', (data,)):
        if exc_info is None:
          assert hashcode is not None, "None from .add of %s" % (S,)
          if ch:
            ch.put(hashcode)
            ch = None
        else:
          e = exc_info[1]
          if isinstance(e, StoreError):
            exc_info = None
          X("================ exc_info=%r", exc_info)
          error("exception from %s.add: %s", S, e, exc_info=exc_info)
          if ok:
            ok = False
            if self.save2:
              # kick off the fallback saves immediately
              X("_BG_ADD: dispatch fallback %r", self.save2)
              fallback = list(self._multicall0(self.save2, 'add_bg', (data,)))
            else:
              error("no fallback Stores")
          continue
      if not ok:
        if fallback:
          failures = []
          for LF, S in fallback:
            hashcode, exc_info = LF.join()
            if exc_info:
              e = exc_info[1]
              if isinstance(e, StoreError):
                exc_info = None
              X("==== ==== ==== exc_info=%r", exc_info)
              error("exception saving to %s: %s", S, exc_info[1], exc_info=exc_info)
              failures.append( (S, e) )
            else:
              if ch:
                ch.put(hashcode)
                ch = None
          if failures:
            raise RuntimeError("exceptions saving to save2: %r" % (failures,))
    finally:
      # mark end of queue
      if ch:
        ch.put(None)
        ch = None
        ch.close()

  def get(self, h):
    ''' Fetch a block from the first Store which has it.
    '''
    with Pfx("%s.get", type(self)):
      for stores in self.read, self.read2:
        for S, data, exc_info in self._multicall(stores, 'get_bg', (h,)):
          with Pfx("%s.get_bg(%s)", S, h):
            if exc_info:
              error("exception", exc_info=exc_info)
            elif data is not None:
              ##XP("got %d bytes", len(data))
              if S not in self.read:
                for copyS in self.copy2:
                  ##XP("copy to %s", copyS)
                  copyS.add_bg(data)
              return data
      return None

  def contains(self, h):
    ''' Test whether the hashcode `h` is in any of the read Stores.
    '''
    for stores in self.read, self.read2:
      for S, result, exc_info in self._multicall(stores, 'contains_bg', (h,)):
        if exc_info:
          error("exception fetching from %s: %s", S, exc_info)
        elif result:
          return True
    return False

  def flush(self):
    ''' Flush all the save Stores.
    '''
    for _ in self._multicall(self.save, 'flush_bg', ()):
      pass

class DataDirStore(MappingStore):
  ''' A MappingStore using a DataDir as its backend.
  '''

  def __init__(
      self,
      name,
      statedirpath,
      *,
      hashclass=None, indexclass=None,
      rollover=None,
      **kw
  ):
    if hashclass is None:
      hashclass = DEFAULT_HASHCLASS
    datadir = DataDir(
        statedirpath, hashclass,
        indexclass=indexclass,
        rollover=rollover)
    MappingStore.__init__(self, name, datadir, **kw)
    self._datadir = datadir

  def startup(self, **kw):
    ''' Startup: open the internal DataDir.
    '''
    super().startup(**kw)
    self._datadir.open()

  def shutdown(self):
    ''' Shutdown: close the internal DataDir.
    '''
    self._datadir.close()
    super().shutdown()

  def init(self):
    ''' Init the supporting data dir.
    '''
    self._datadir.init()

  def localpathto(self, rpath):
    ''' Compute the full path from a relative path.
    '''
    return self._datadir.localpathto(rpath)

def PlatonicStore(name, statedirpath, *a, meta_store=None, **kw):
  ''' Factory function for platonic Stores.
      This is needed because if a meta_store is specified then it
      must be included as a block source in addition to the core
      platonic Store.
  '''
  if meta_store is None:
    return _PlatonicStore(name, statedirpath, *a, **kw)
  PS = _PlatonicStore(name, statedirpath, *a, meta_store=meta_store, **kw)
  S = ProxyStore(
      name,
      save=(),
      read=(PS, meta_store)
  )
  S.get_Archive = PS.get_Archive
  return S

class _PlatonicStore(MappingStore):
  ''' A MappingStore using a PlatonicDir as its backend.
  '''

  def __init__(
      self, name, statedirpath,
      *,
      hashclass=None, indexclass=None,
      follow_symlinks=False, archive=None, meta_store=None,
      flag_prefix=None,
      runstate=None,
      **kw
  ):
    if hashclass is None:
      hashclass = DEFAULT_HASHCLASS
    datadir = PlatonicDir(
        statedirpath, hashclass,
        follow_symlinks=follow_symlinks,
        archive=archive, meta_store=meta_store,
        flag_prefix=flag_prefix,
    )
    MappingStore.__init__(self, name, datadir, **kw)
    self._datadir = datadir
    self.readonly = True

  def startup(self, **kw):
    ''' Startup: open the internal DataDir.
    '''
    super().startup(**kw)
    self._datadir.open()

  def shutdown(self):
    ''' Shutdown: close the internal DataDir.
    '''
    self._datadir.close()
    super().shutdown()

  def get_Archive(self, archive_name=None):
    ''' PlatonicStore Archive are stored in the internal DataDir.
    '''
    return self._datadir.get_Archive(archive_name)

class _ProgressStoreTemplateMapping(object):

  def __init__(self, PS):
    self.PS = PS

  def __getitem__(self, key):
    try:
      category, aspect = key.rsplit('_', 1)
    except ValueError:
      category = key
      aspect = 'position'
    P = self.PS._progress[category]
    try:
      value = getattr(P, aspect)
    except AttributeError as e:
      raise KeyError("%s: aspect=%r: %s" % (key, aspect, e))
    return value

class ProgressStore(BasicStoreSync):
  ''' A shim for another Store to do progress reporting.
      TODO: planning to redo basic store methods as shims, with
      implementations supplying _foo methods across the board
      instead.
  '''

  def __init__(
      self,
      name, S,
      template='rq  {requests_position}  {requests_throughput}/s',
      **kw
  ):
    ''' Wrapper for a Store which collects statistics on use.
    '''
    lock = kw.pop('lock', None)
    if lock is None:
      lock = S._lock
    BasicStoreAsync.__init__(self, "ProgressStore(%s)" % (name,), lock=lock, **kw)
    self.S = S
    self.template = template
    self.template_mapping = _ProgressStoreTemplateMapping(self)
    Ps = {}
    for category in 'requests', \
                    'adds', 'gets', 'contains', 'flushes', \
                    'bytes_stored', 'bytes_fetched':
      Ps[category] = Progress(name='-'.join((str(S), category)), throughput_window=4)
    self._progress = Ps

  def __str__(self):
    return self.status_text()

  def startup(self):
    super().startup()
    self.S.open()

  def shutdown(self):
    self.S.close()
    super().shutdown()

  def status_text(self, template=None):
    ''' Return a status text utilising the progress statistics.
    '''
    if template is None:
      template = self.template
    return template.format_map(self.template_mapping)

  def add(self, data):
    progress = self._progress
    progress['requests'] += 1
    size = len(data)
    LF = self.S.add_bg(data)
    del data
    progress['adds'] += 1
    progress['bytes_stored'] += size
    return LF()

  def get(self, h):
    progress = self._progress
    progress['requests'] += 1
    LF = self.S.get_bg(h)
    progress['gets'] += 1
    data = LF()
    progress['bytes_fetched'] += len(data)
    return data

  def contains(self, h):
    progress = self._progress
    progress['requests'] += 1
    LF = self.S.contains_bg(h)
    progress['contains'] += 1
    return LF()

  def flush(self):
    progress = self._progress
    progress['requests'] += 1
    LF = self.S.flush_bg()
    progress['flushes'] += 1
    return LF()

  @property
  def requests(self):
    ''' The number of requests.
    '''
    return self._progress['requests'].position

if __name__ == '__main__':
  from .store_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Parsers for data streams, emitting data and offsets.
# These sit in front of the core rolling hash blockifier.
#   - Cameron Simpson <cs@cskk.id.au> 05mar2017
#

''' Parsers for various recognised data formats to aid block edge selection.
'''

from functools import partial
from os.path import basename, splitext
from cs.buffer import chunky
from cs.logutils import warning, exception
from cs.pfx import Pfx, PfxThread
from cs.queues import IterableQueue
from .datafile import DataFileReader

def linesof(chunks):
  ''' Process binary chunks, yield binary lines ending in '\n'.
      The final line might not have a trailing newline.
  '''
  pending = []
  for chunk in chunks:
    # get a memoryview so that we can cheaply queue bits of it
    mv_chunk = memoryview(chunk)
    upto = 0
    # but scan the chunk, because memoryviews do not have .find
    nlpos = chunk.find(b'\n')
    while nlpos >= 0:
      pending.append(mv_chunk[upto:nlpos+1])
      yield b''.join(pending)
      pending = []
      upto = nlpos + 1
      nlpos = chunk.find(b'\n', upto)
    # stash incomplete line in pending
    if upto < len(chunk):
      pending.append(mv_chunk[upto:])
  if pending:
    yield b''.join(pending)

def scan_text(bfr, prefixes=None):
  ''' Scan textual data, yielding offsets of lines starting with
      useful prefixes, such as function definitions.
  '''
  with Pfx("scan_text"):
    if prefixes is None:
      prefixes = PREFIXES_ALL
    prefixes = [
        (
            prefix
            if isinstance(prefix, bytes)
            else (
                bytes(prefix)
                if isinstance(prefix, memoryview)
                else (
                    prefix.encode('utf-8')
                    if isinstance(prefix, str)
                    else prefix
                )
            )
        ) for prefix in prefixes
    ]
    offset = 0
    for line in linesof(bfr):
      next_offset = None
      for prefix in prefixes:
        if line.startswith(prefix):
          next_offset = offset
          break
      if next_offset is not None:
        yield next_offset
      offset += len(line)

scan_text_from_chunks = chunky(scan_text)

def report_offsets(bfr, run_parser):
  ''' Dispatch a parser in a separate Thread, return an IterableQueue yielding offsets.
      `bfr`: a CornuCopyBuffer providing data to parse
      `run_parser`: a callable which runs the parser; it should accept a
        CornuCopyBuffer as its sole argument.
      This function allocates an IterableQueue to receive the parser offset
      reports and sets the CornuCopyBuffer with report_offset copying
      offsets to the queue.
      It is the task of the parser to call `bfr.report_offset` as
      necessary to indicate suitable offsets.
  '''
  with Pfx("report_offsets(bfr,run_parser=%s)", run_parser):
    offsetQ = IterableQueue()
    if bfr.copy_offsets is not None:
      warning("bfr %s already has copy_offsets, replacing", bfr)
    bfr.copy_offsets = offsetQ.put
    def thread_body():
      with Pfx("parser-thread"):
        try:
          run_parser(bfr)
        except Exception as e:
          exception("exception: %s", e)
          raise
        finally:
          offsetQ.close()
    T = PfxThread(target=thread_body)
    T.start()
    return offsetQ

report_offsets_from_chunks = chunky(report_offsets)

def scan_vtd(bfr):
  ''' Scan a datafile from `bfr` and yield chunk start offsets.
  '''
  with Pfx("scan_vtd"):
    def run_parser(bfr):
      for offset, *_ in DataFileReader.scanbuffer(bfr):
        bfr.report_offset(offset)
    return report_offsets(bfr, run_parser)

def scan_mp3(bfr):
  ''' Scan MP3 data from `bfr` and yield frame start offsets.
  '''
  from cs.mp3 import framesof as parse_mp3_from_buffer
  with Pfx("scan_mp3"):
    def run_parser(bfr):
      for _ in parse_mp3_from_buffer(bfr):
        pass
    return report_offsets(bfr, run_parser)

scan_mp3_from_chunks = chunky(scan_mp3)

def scan_mp4(bfr):
  ''' Scan ISO14496 input and yield Box start offsets.
  '''
  from cs.iso14496 import parse_buffer as parse_mp4_from_buffer
  with Pfx("parse_mp4"):
    def run_parser(bfr):
      for _ in parse_mp4_from_buffer(bfr, discard_data=True):
        pass
    return report_offsets(bfr, run_parser)

parse_mp4_from_chunks = chunky(scan_mp4)

def scanner_from_filename(filename):
  ''' Choose a scanner based a filename.
      Returns None if these is no special scanner.
  '''
  _, ext = splitext(basename(filename))
  if ext:
    assert ext.startswith('.')
    parser = SCANNERS_BY_EXT.get(ext[1:].lower())
    if parser is not None:
      return parser
  return None

def scanner_from_mime_type(mime_type):
  ''' Choose a scanner based a mime_type.
  '''
  return SCANNERS_BY_MIME_TYPE.get(mime_type)

PREFIXES_MAIL = ( 'From ', '--' )
PREFIXES_PYTHON = (
    'def ', '  def ', '    def ', '\tdef ',
    'class ', '  class ', '    class ', '\tclass ',
)
PREFIXES_GO = (
    'func ',
)
PREFIXES_PERL = (
    'package ', 'sub ',
)
PREFIXES_PDF = (
    '<<',
    'stream',
)
PREFIXES_SH = (
    'function ',
)
PREFIXES_SQL_DUMP = (
    'INSERT INTO ',
    'DROP TABLE ',
    'CREATE TABLE ',
)

SCANNERS_BY_EXT = {
    'go': partial(scan_text, prefixes=PREFIXES_GO),
    'mp3': scan_mp3,
    'mp4': scan_mp4,
    'pdf': partial(scan_text, prefixes=PREFIXES_PDF),
    'pl': partial(scan_text, prefixes=PREFIXES_PERL),
    'pm': partial(scan_text, prefixes=PREFIXES_PERL),
    'py': partial(scan_text, prefixes=PREFIXES_PYTHON),
    'sh': partial(scan_text, prefixes=PREFIXES_SH),
    'sql': partial(scan_text, prefixes=PREFIXES_SQL_DUMP),
    'vtd': scan_vtd,
}

SCANNERS_BY_MIME_TYPE = {
    'text/x-go': partial(scan_text, prefixes=PREFIXES_GO),
    'audio/mpeg': scan_mp3,
    'video/mp4': scan_mp4,
    'text/x-perl': partial(scan_text, prefixes=PREFIXES_PERL),
    'text/x-python': partial(scan_text, prefixes=PREFIXES_PYTHON),
    'text/x-sh': partial(scan_text, prefixes=PREFIXES_SH),
}

PREFIXES_ALL = (
    PREFIXES_MAIL
    + PREFIXES_PYTHON
    + PREFIXES_GO
    + PREFIXES_PERL
    + PREFIXES_SH
    + PREFIXES_SQL_DUMP
)



#!/usr/bin/env python3
#
# Functions for running a server.
# - Cameron Simpson <cs@cskk.id.au> 30may2018
#

''' Server stub functions.
'''

from .socket import TCPStoreServer, UNIXSocketStoreServer

def serve_tcp(bind_addr, exports=None, runstate=None):
  ''' Return an open TCPStoreServer.
  '''
  srv = TCPStoreServer(bind_addr, exports=exports, runstate=runstate)
  srv.open()
  return srv

def serve_socket(socket_path, exports=None, runstate=None):
  ''' Return an open UNIXSocketStoreServer.
  '''
  srv = UNIXSocketStoreServer(socket_path, exports=exports, runstate=runstate)
  srv.open()
  return srv



#!/usr/bin/python3
#
# Map block to leaves and offsets.
# - Cameron Simpson <cs@cskk.id.au> 07feb2018
#

'''
A flat index of leaf offsets and their hashcodes to speed data
lookup from an indirect Block. This produces memory mapped indices
to bypass the need to walk the block tree to fetch leaf data.
'''

from bisect import bisect_right
from collections import namedtuple
from mmap import mmap, MAP_PRIVATE, PROT_READ
import os
from os.path import isdir, exists as pathexists, join as joinpath
from struct import Struct
import sys
from tempfile import TemporaryFile, NamedTemporaryFile
from threading import Thread
from cs.excutils import logexc
from cs.logutils import warning
from cs.pfx import Pfx
from cs.py.func import prop
from cs.resources import RunStateMixin
from cs.x import X
from . import defaults
from .block import HashCodeBlock

# The record format uses 4 byte integer offsets
# so this is the maximum (and default) scale for the memory maps.
OFFSET_SCALE = 2 ** 32

OFF_STRUCT = Struct('<L')

_MapEntry = namedtuple('MapEntry', 'index offset span hashcode')
class MapEntry(_MapEntry):
  ''' A blockmap entry (index, offset, span, hashcode) and related properties.
  '''

  @prop
  def leaf(self):
    ''' Return the leaf block for this entry.
    '''
    return HashCodeBlock(hashcode=self.hashcode, span=self.span)

  @prop
  def data(self):
    ''' Return the data from this leaf block.
    '''
    return self.leaf.data

class MappedFD:
  ''' Manage a memory map of the contents of a file
      representing a block's backing leaf content.

      The file contains records (offset, hashcode) in offset order,
      being the starting offset of a leaf block relative to the
      start of the map and the leaf block's hashcode.

      Aside from the first map, the first record need not have
      offset=0 because the leading bytes in this submap probably
      come from a leaf block from a preceeding submap that overflows
      into this map's range.
  '''

  def __init__(self, f, hashclass):
    ''' Initialise a MappedFD from a file.
        `f`: the file whose contents will be mapped
          This may be an open file object or the path to a persistent map file.
          The file is expected to be prefilled with complete records.
        `hashclass`: the type of hashcodes stored in the map, used
          for sizing and for returning this type from the entry bytes
        If `f` is a file path it is opened for read.
        If `f` is an open file, the file's file descriptor is dup()ed
        and the dup used to manage the memory map, allowing the
        original file to be closed by the caller.
    '''
    self.hashclass = hashclass
    self.rec_size = OFF_STRUCT.size + hashclass.HASHLEN
    if isinstance(f, str):
      with Pfx("open(%r)", f):
        fd = os.open(f, os.O_RDONLY)
    else:
      # `f` should be an open file
      f.flush()
      with Pfx("dup(%r.fileno())", f):
        fd = os.dup(f.fileno())
    self.fd = fd
    self.mapped = mmap(fd, 0, flags=MAP_PRIVATE, prot=PROT_READ)
    self.record_count = self.mapped.size() // self.rec_size
    assert self.mapped.size() % self.rec_size == 0, \
        "mapped.size()=%s, rec_size=%s, modulus=%d" \
        % (self.mapped.size(), self.rec_size, self.mapped.size() % self.rec_size)

  def __del__(self):
    ''' Release resouces on object deletion.
    '''
    self.close()

  def close(self):
    ''' Release the map and its file descriptor.
    '''
    self.mapped.close()

  def __len__(self):
    return self.record_count

  def locate(self, offset):
    ''' Locate and return the MapEntry containing the specified `offset`.

        If the offset is not contained within this map then the returned
        MapEntry will contain (index=-1, offset=None, span=None, hashcode=None).
    '''
    if offset < 0 or offset >= OFFSET_SCALE:
      raise ValueError("offset(%s) out of range 0:%s" % (offset, OFFSET_SCALE))
    i = bisect_right(self, offset)
    assert 0 < i <= len(self)
    i -= 1
    entry = self.entry(i)
    if offset < entry.offset or offset >= entry.offset + entry.span:
      X("submap.locate(offset=%d): entry=%s OUT OF RANGE", offset, entry)
      entry = MapEntry(-1, None, None, None)
    return entry

  def __getitem__(self, i):
    ''' Return the offset for entry(i), to support bisect.
    '''
    try:
      offset = self.offset(i)
    except ValueError:
      raise IndexError(i)
    return offset

  def entries(self, i):
    ''' Yield MapEntry instances starting at index `i`.
    '''
    record_count = self.record_count
    if i < 0 or i >= record_count:
      raise ValueError("index(%s) out of range 0:%d" % (i, record_count))
    while i < self.record_count:
      yield self.entry(i)
      i += 1

  def entry(self, i):
    ''' Fetch the MapEntry at index `i`.
    '''
    i0 = i
    if i < 0:
      # convert negative index
      i = self.record_count + i
    if i < 0 or i >= self.record_count:
      raise ValueError(i0)
    mapped = self.mapped
    rec_size = self.rec_size
    rec_offset = i * rec_size
    hash_offset = rec_offset + OFF_STRUCT.size
    next_rec_offset = rec_offset + rec_size
    offset, = OFF_STRUCT.unpack(mapped[rec_offset:hash_offset])
    hashcode = self.hashclass.from_hashbytes(mapped[hash_offset:next_rec_offset])
    if i == self.record_count - 1:
      span = len(defaults.S[hashcode])
    else:
      next_hash_offset = next_rec_offset + OFF_STRUCT.size
      next_offset, = OFF_STRUCT.unpack(mapped[next_rec_offset:next_hash_offset])
      span = next_offset - offset
    return MapEntry(i, offset, span, hashcode)

  def offset(self, i):
    ''' Fetch the offset of index `i`.
    '''
    i0 = i
    if i < 0:
      # convert negative index
      i = self.record_count + i
    if i < 0 or i >= self.record_count:
      raise ValueError(i0)
    mapped = self.mapped
    rec_size = self.rec_size
    rec_offset = i * rec_size
    hash_offset = rec_offset + OFF_STRUCT.size
    offset, = OFF_STRUCT.unpack(mapped[rec_offset:hash_offset])
    return offset

  @prop
  def start(self):
    ''' The offset of the first leaf in the mapping.
    '''
    return self.offset(0)

class BlockMap(RunStateMixin):
  ''' A fast mapping of offsets to leaf block hashcodes.
  '''

  def __init__(self, block, mapsize=None, blockmapdir=None):
    ''' Initialise the BlockMap, dispatch the index generator.
        `block`: the source Block
        `mapsize`: the size of each index map, default `OFFSET_SCALE`
        `blockmapdir`: the pathname for persistent storage of BlockMaps
    '''
    if mapsize is None:
      mapsize = OFFSET_SCALE
    elif mapsize <= 0 or mapsize > OFFSET_SCALE:
      raise ValueError(
          "mapsize(%d) out of range, must be >0 and <=%d"
          % (mapsize, OFFSET_SCALE))
    # DEBUGGING
    if blockmapdir is None:
      blockmapdir = '/Users/cameron/hg/css-venti/test_blockmaps'
      X("BlockMap: set blockmapdir to %r (was None)", blockmapdir)
    else:
      X("BlockMap: supplied blockmapdir=%r", blockmapdir)
    RunStateMixin.__init__(self, "BlockMap")
    from .block import _IndirectBlock
    if not isinstance(block, _IndirectBlock):
      raise TypeError("block needs to be a _IndirectBlock, got a %s instead" % (type(block),))
    hashcode = block.superblock.hashcode
    hashclass = type(hashcode)
    self.hashclass = hashclass
    self.mapsize = mapsize
    if blockmapdir is None:
      self.mappath = mappath = None
    else:
      self.mappath = mappath = joinpath(blockmapdir, "mapsize:%d" % (mapsize,), hashcode.filename)
      if not isdir(mappath):
        with Pfx("makedirs(%r)", mappath):
          X("MKDIR %r", mappath)
          os.makedirs(mappath)
    self.block = block
    self.S = defaults.S
    nsubmaps = len(block) // mapsize + 1
    submaps = [None] * nsubmaps
    self.maps = submaps
    mapped_to = 0
    self.rec_size = OFF_STRUCT.size + len(hashcode)
    self._loaded = False
    # preattach any existing blockmap files
    if mappath is not None:
      for submap_index in range(nsubmaps):
        submappath = joinpath(mappath, '%d.blockmap' % (submap_index,))
        if not pathexists(submappath):
          break
        # existing map, attach and install, advance and restart loop
        X("Blockmap.__init__: preattach existing map %r", submappath)
        submaps[submap_index] = MappedFD(submappath, hashclass)
        mapped_to += mapsize
    self.mapped_to = mapped_to
    if mapped_to < len(block):
      X("BlockMap.__init__: dispatch worker to scan from offset %d ...", mapped_to)
      self._worker = Thread(target=self._load_maps, args=(defaults.S,))
      self.runstate.start()
      self._worker.start()
    else:
      self._worker = None

  def join(self):
    ''' Wait for the worker to complete.
    '''
    self.runstate.cancel()
    worker = self._worker
    if worker is not None:
      self._worker.join()

  def __del__(self):
    ''' Release resources on object deletion.
    '''
    self.close()

  def close(self):
    ''' Release the resources associated with the BlockMap.
    '''
    X("BlockMap.close...")
    self.cancel()
    self.join()
    maps = self.maps
    for i in range(len(maps)-1):
      submap = maps[i]
      if submap is not None:
        submap.close()
        maps[i] = None

  @logexc
  def _load_maps(self, S):
    ''' Load leaf offsets and hashcodes into the unfilled portion of the blockmap.
    '''
    offset = self.mapped_to
    mapsize = self.mapsize
    submap_index = offset // mapsize - 1
    with S:
      runstate = self.runstate
      block = self.block
      blocklen = len(block)
      hashclass = self.hashclass
      submaps = self.maps
      submap_fp = None
      submap_path = None
      nleaves = 0
      while offset < blocklen and not runstate.cancelled:
        for leaf, start, length in block.slices(offset, len(block)):
          if runstate.cancelled:
            break
          if start > 0:
            # partial block, skip to next
            offset += length
            continue
          leaf_submap_index = offset // mapsize
          leaf_submap_offset = offset % mapsize
          if submap_index < leaf_submap_index:
            # this leaf belongs in a new submap
            if submap_fp is not None:
              # consume the submap in progress
              submap = MappedFD(submap_fp, hashclass)
              if submap_path is not None:
                X("NEW submap: %r", submap_path)
                os.link(submap_fp.name, submap_path)
              submaps[submap_index] = submap
              last_entry = submap.entry(-1)
              self.mapped_to = last_entry.offset + last_entry.span
              submap_fp.close()
              submap_fp = None
            # advance to the correct submap index
            submap_index = leaf_submap_index
            if self.mappath:
              # if we're doing persistent submaps...
              submap_path = joinpath(self.mappath, '%d.blockmap' % (submap_index,))
              if pathexists(submap_path):
                # existing map, attach and install, advance and restart loop
                X("LOAD MAPS: attach existing map %r", submap_path)
                submaps[submap_index] = MappedFD(submap_path, hashclass)
                offset = ( submap_index + 1 ) * mapsize
                X("LOAD MAPS: skip to offset=0x%x", offset)
                break
              # start a new persistent file to attach later
              submap_fp = NamedTemporaryFile('wb')
            else:
              # not persistent - start a new temp file to attach later
              submap_fp = TemporaryFile('wb')
              submap_path = None
          # post condition: correct submap_index and correct submap_fp state
          assert (
              submap_index == leaf_submap_index
              and submap_index < len(submaps)
              and (
                  submap_fp is None
                  if self.mappath and pathexists(submap_path)
                  else submap_fp is not None
              )
          )
          try:
            h = leaf.hashcode
          except AttributeError:
            # make a conventional HashCodeBlock and index that
            data = leaf.data
            if len(data) >= 65536:
              warning("promoting %d bytes from %s to a new HashCodeBlock", len(data), leaf)
            leaf = HashCodeBlock(data=data)
            h = leaf.hashcode
          submap_fp.write(OFF_STRUCT.pack(leaf_submap_offset))
          submap_fp.write(h)
          offset += leaf.span
          nleaves += 1
          if nleaves % 4096 == 0:
            X(
                "LOAD MAPS: processed %d leaves in %gs (%d leaves/s)",
                nleaves, runstate.run_time, nleaves // runstate.run_time)
      X("LOAD MAPS: leaf scan finished")
      # attach final submap after the loop if one is in progress
      if submap_fp is not None:
        # consume the submap in progress
        submap = MappedFD(submap_fp, hashclass)
        if submap_path is not None:
          X("NEW submap: %r", submap_path)
          os.link(submap_fp.name, submap_path)
        submaps[submap_index] = submap
        last_entry = submap.entry(-1)
        self.mapped_to = last_entry.offset + last_entry.span
        submap_fp.close()
        submap_fp = None
      X("LOAD MAPS: COMPLETE")

  def self_check(self):
    ''' Perform some integrity tests.
    '''
    ##assert self._loaded
    for i, submap in enumerate(self.maps):
      if submap is None:
        continue
      assert isinstance(submap, MappedFD), \
          "maps[%d] is not a MappedFD: %r; maps=%r" % (i, type(submap), self.maps)

  def datafrom(self, offset=0, span=None):
    ''' Generator yielding data from [offset:offset+span] from the relevant leaves.

        Parameters:
        * `offset`: starting offset within `self.block`, default `0`
        * `span`: number of bytes to cover; if omitted or None, the
          span runs to the end of `self.block`
    '''
    for leaf, start, end in self.slices(offset, span):
      assert start < end
      assert start >= 0
      assert end <= len(leaf)
      yield leaf[start:end]

  # TODO: accept start,end instead of start,span like other slices methods
  def slices(self, offset, span=None):
    ''' Generator yielding (leaf, start, end) from [offset:offset+span].

        Parameters:
        `offset`: starting offset within `self.block`
        `span`: number of bytes to cover; if omitted or None, the
          span runs to the end of self.block
    '''
    if span is None:
      span = len(self.block) - offset
    if offset < 0:
      raise ValueError("offset(%d) should be >= 0" % (offset,))
    if span < 0:
      raise ValueError("span(%d) should be >= 0" % (span,))
    if span == 0:
      return
    maps = self.maps
    mapsize = self.mapsize
    while span > 0:
      if self.mapped_to <= offset:
        # outside the mapped range
        # use the normal Block.slices method without any blockmap
        yield from self.block.slices(offset, offset + span, no_blockmap=True)
        return
      # we can get the start of the span from the blockmap
      submap_index = offset // mapsize
      # locate starting entry
      submap = maps[submap_index]
      if submap is None or submap.start > offset:
        # leaf must come from the preceeding submap
        while submap is None or submap.start > offset:
          submap_index -= 1
          submap = maps[submap_index]
          if submap:
            break
        entry = submap.entry(-1)
      else:
        submap_offset = offset % mapsize
        entry = submap.locate(submap_offset)
      submap_base = submap_index * mapsize
      while span > 0 and entry:
        entry_offset = submap_base + entry.offset
        assert entry_offset <= offset < entry_offset + entry.span
        leaf = entry.leaf
        leaf_start = offset - entry_offset
        leaf_subspan = min(entry.span - leaf_start, span)
        leaf_end = leaf_start + leaf_subspan
        yield leaf, leaf_start, leaf_end
        offset += leaf_subspan
        span -= leaf_subspan
        if span == 0:
          break
        if entry.index == submap.record_count - 1:
          break
        # fetch next entry
        entry = submap.entry(entry.index + 1)

  def data(self, offset, span):
    ''' Return the data from [offset:offset+span] as a single bytes object.
    '''
    return b''.join(self.datafrom(offset, span))

  def __getitem__(self, index):
    ''' Return a single byte from the BlockMap.
    '''
    if isinstance(index, int):
      return next(self.datafrom(index))[0]
    if index.step is not None and index.step != 1:
      raise ValueError("invalid slice: step=%s" % (index.step,))
    start = 0 if index.start is None else index.start
    span = None if index.stop is None else index.stop - start
    if span < 0:
      raise ValueError("invalid span: stop(%s) < start(%s)" % (index.stop, index.start))
    if span == 0:
      return b''
    return b''.join(self.datafrom(start, span))

if __name__ == '__main__':
  from .blockmap_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Configuration file and services.
#   - Cameron Simpson <cs@cskk.id.au> 2017-12-25
#

'''
Store definition configuration file.
'''

from configparser import ConfigParser
import os
from os.path import abspath, isabs as isabspath, join as joinpath, exists as pathexists
import sys
from cs.fileutils import shortpath, longpath
from cs.logutils import debug, warning, error
from cs.pfx import Pfx
from cs.result import Result
from . import Lock, DEFAULT_CONFIG
from .archive import Archive
from .cache import FileCacheStore, MemoryCacheStore
from .compose import parse_store_specs
from .convert import get_integer, \
    convert_param_int, convert_param_scaled_int, \
    convert_param_path
from .store import PlatonicStore, ProxyStore, DataDirStore
from .socket import TCPClientStore, UNIXSocketClientStore

def Store(spec, config, runstate=None):
  ''' Factory to construct Stores from string specifications.
  '''
  return config.Store_from_spec(spec, runstate=runstate)

class Config:
  ''' A configuration specification.

      This can be driven by any mapping of mappings: {clause_name => {param => value}}.
      It is modelled on a .ini file, with named clauses each containing named values.

      Parameters:
      * `config_map`: either a mapping of mappings: `{clause_name: {param: value}}`
        or the filename of a file in `.ini` format
      * `environ`: optional environment mapp for `$varname` substitution.
        Default: `os.environ`
  '''

  def __init__(self, config_map, environ=None, default_config=None):
    if environ is None:
      environ = os.environ
    if default_config is None:
      default_config = DEFAULT_CONFIG
    self.environ = environ
    config = ConfigParser()
    if isinstance(config_map, str):
      self.path = path = config_map
      with Pfx(path):
        read_ok = False
        if pathexists(path):
          try:
            config.read(path)
          except OSError as e:
            error("read error: %s", e)
          else:
            read_ok = True
        else:
          warning("missing config file")
      if not read_ok:
        warning("falling back to default configuration")
        config.read_dict(default_config)
    else:
      config.read_dict(config_map)
    self.map = config
    self._stores_by_name = {}  # clause_name => Result->Store
    self._lock = Lock()

  def __str__(self):
    if self.path is None:
      return repr(self)
    return "Config(%s)" % (shortpath(self.path),)

  def write(self, fp=None):
    ''' Write the configuration out to the file `fp`.
    '''
    if fp is None:
      fp = sys.stdout
    self.map.write(fp)

  def __getitem__(self, clause_name):
    ''' Return the Store defined by the named clause.
    '''
    with self._lock:
      R = self._stores_by_name.get(clause_name)
      if R is not None:
        return R()
      R = self._stores_by_name[clause_name] = Result("[%s]" % (clause_name,))
    # not previously accessed, construct S
    store_name = "%s[%s]" % (self, clause_name)
    with Pfx(store_name):
      clause = dict(self.map[clause_name])
      for discard in 'address', :
        clause.pop(discard, None)
      try:
        store_type = clause.pop('type')
      except KeyError:
        raise ValueError("missing type field in clause")
      S = self.new_Store(
          store_name,
          store_type,
          clause,
          clause_name=clause_name
      )
      R.result = S
    return S

  def get_default(self, param, default=None):
    ''' Fetch a default parameter from the [GLOBALS] clause.
    '''
    G = self.map['GLOBAL']
    if not G:
      return default
    return G.get(param, default)

  def get_clause(self, clause_name):
    ''' Return the clause without opening it as a Store.
    '''
    return self.map[clause_name]

  @property
  def basedir(self):
    ''' The default location for local archives and stores.
    '''
    return longpath(self.get_default('basedir'))

  @property
  def mountdir(self):
    ''' The default directory for mount points.
    '''
    return longpath(self.get_default('mountdir'))

  def archive(self, archivename):
    ''' Return the Archive named `archivename`.
    '''
    if (
        not archivename
        or '.' in archivename
        or '/' in archivename
    ):
      raise ValueError("invalid archive name: %r" % (archivename,))
    arpath = joinpath(self.basedir, archivename + '.vt')
    return Archive(arpath)

  def Store_from_spec(self, store_spec, runstate=None):
    ''' Factory function to return an appropriate BasicStore* subclass
        based on its argument:

          store:...       A sequence of stores. Save new data to the
                          first, seek data in all from left to right.

        Multiple stores are combined into a ProxyStore which saves
        to the first Store and reads from all the Stores.

        See also the .Stores_from_spec method, which returns the
        separate Stores unassembled.
    '''
    with Pfx(repr(store_spec)):
      stores = self.Stores_from_spec(store_spec)
      if not stores:
        raise ValueError("empty Store specification: %r" % (store_spec,))
      if len(stores) == 1:
        S = stores[0]
      else:
        # multiple stores: save to the front store, read first from the
        # front store then from the rest
        S = ProxyStore(store_spec, stores[0:1], stores[0:1], read2=stores[1:])
      if runstate is not None:
        S.runstate = runstate
      return S

  def Stores_from_spec(self, store_spec):
    ''' Parse a colon separated list of Store specifications,
        return a list of Stores.
    '''
    store_specs = list(parse_store_specs(store_spec))
    if not store_specs:
      raise ValueError("empty Store specification: %r" % (store_specs,))
    stores = [
        self.new_Store(store_text, store_type, params)
        for store_text, store_type, params
        in store_specs
    ]
    return stores

  def new_Store(self, store_name, store_type, params, clause_name=None):
    ''' Construct a store given its specification.
    '''
    with Pfx("new_Store(%r,type=%r,params=%r,...)", store_name, store_type, params):
      if not isinstance(params, dict):
        params = dict(params)
      # process general purpose params
      # blockmapdir: location to store persistent blockmaps
      blockmapdir = params.pop('blockmapdir', None)
      if store_name is None:
        store_name = str(self) + '[' + clause_name + ']'
      if store_type == 'config':
        S = self.config_Store(store_name, **params)
      elif store_type == 'datadir':
        S = self.datadir_Store(store_name, clause_name, **params)
      elif store_type == 'filecache':
        convert_param_int(params, 'max_files')
        convert_param_scaled_int(params, 'max_file_size')
        S = self.filecache_Store(store_name, clause_name, **params)
      elif store_type == 'memory':
        convert_param_scaled_int(params, 'max_data')
        S = self.memory_Store(store_name, clause_name, **params)
      elif store_type == 'platonic':
        S = self.platonic_Store(store_name, clause_name, **params)
      elif store_type == 'proxy':
        S = self.proxy_Store(store_name, **params)
      elif store_type == 'socket':
        if 'socket_path' not in params:
          params['socket_path'] = clause_name
        convert_param_path(params, 'socket_path')
        S = self.socket_Store(store_name, **params)
      elif store_type == 'tcp':
        if 'host' not in params:
          params['host'] = clause_name
        S = self.tcp_Store(store_name, **params)
      else:
        raise ValueError("unsupported type %r" % (store_type,))
      if S.config is None:
        S.config = self
      if blockmapdir is not None:
        S.blockmapdir = blockmapdir
      return S

  def config_Store(
      self,
      _,    # store_name, unused
      *,
      type_=None,
      clause_name=None,
  ):
    ''' Construct a Store from a reference to a configuration clause.
    '''
    if type_ is None:
      type_ = 'config'
    else:
      assert type_ == 'config'
    if clause_name is None:
      raise ValueError("clause_name may not be None")
    return self[clause_name]

  def datadir_Store(
      self,
      store_name, clause_name,
      *,
      type_=None,
      path=None,
      basedir=None,
  ):
    ''' Construct a DataDirStore from a "datadir" clause.
    '''
    if type_ is not None:
      assert type_ == 'datadir'
    if basedir is None:
      basedir = self.get_default('basedir')
    if path is None:
      path = clause_name
    path = longpath(path)
    if not isabspath(path):
      if path.startswith('./'):
        path = abspath(path)
      else:
        if basedir is None:
          raise ValueError('relative path %r but no basedir' % (path,))
        basedir = longpath(basedir)
        path = joinpath(basedir, path)
    return DataDirStore(store_name, path)

  def filecache_Store(
      self,
      store_name, clause_name,
      *,
      type_=None,
      path=None,
      max_files=None,
      max_file_size=None,
      basedir=None,
      backend=None,
  ):
    ''' Construct a FileCacheStore from a "filecache" clause.
    '''
    if type_ is not None:
      assert type_ == 'filecache'
    if basedir is None:
      basedir = self.get_default('basedir')
    if path is None:
      path = clause_name
      debug("path from clausename: %r", path)
    path = longpath(path)
    debug("longpath(path) ==> %r", path)
    if backend is None:
      backend_store = None
    else:
      backend_store = self.Store_from_spec(backend)
    if not isabspath(path):
      if path.startswith('./'):
        path = abspath(path)
        debug("abspath ==> %r", path)
      else:
        if basedir is None:
          raise ValueError('relative path %r but no basedir' % (path,))
        basedir = longpath(basedir)
        debug("longpath(basedir) ==> %r", basedir)
        path = joinpath(basedir, path)
        debug("path ==> %r", path)
    return FileCacheStore(
        store_name, backend_store, path,
        max_cachefile_size=max_file_size,
        max_cachefiles=max_files,
    )

  def memory_Store(
      self,
      store_name, clause_name,
      *,
      type_=None,
      max_data=None,
  ):
    ''' Construct a PlatonicStore from a "datadir" clause.
    '''
    if type_ is not None:
      assert type_ == 'memory'
    if max_data is None:
      raise ValueError("missing max_data")
    return MemoryCacheStore(store_name, max_data)

  def platonic_Store(
      self,
      store_name, clause_name,
      *,
      type_=None,
      path=None,
      basedir=None,
      follow_symlinks=False,
      meta=None,
      archive=None,
  ):
    ''' Construct a PlatonicStore from a "datadir" clause.
    '''
    if type_ is not None:
      assert type_ == 'platonic'
    if basedir is None:
      basedir = self.get_default('basedir')
    if path is None:
      path = clause_name
      debug("path from clausename: %r", path)
    path = longpath(path)
    debug("longpath(path) ==> %r", path)
    if not isabspath(path):
      if path.startswith('./'):
        path = abspath(path)
        debug("abspath ==> %r", path)
      else:
        if basedir is None:
          raise ValueError('relative path %r but no basedir' % (path,))
        basedir = longpath(basedir)
        debug("longpath(basedir) ==> %r", basedir)
        path = joinpath(basedir, path)
        debug("path ==> %r", path)
    if follow_symlinks is None:
      follow_symlinks = False
    if meta is None:
      meta_store = None
    elif isinstance(meta, str):
      meta_store = Store(meta, self)
    if isinstance(archive, str):
      archive = longpath(archive)
    return PlatonicStore(
        store_name, path,
        hashclass=None, indexclass=None,
        follow_symlinks=follow_symlinks,
        meta_store=meta_store, archive=archive,
        flag_prefix='VT_' + clause_name,
    )

  def proxy_Store(
      self,
      store_name,
      *,
      type_=None,
      save=None,
      read=None,
      save2=None,
      read2=None,
      copy2=None,
  ):
    ''' Construct a ProxyStore.
    '''
    if type_ is not None:
      assert type_ == 'proxy'
    if save is None:
      save_stores = []
      readonly = True
    else:
      if isinstance(save, str):
        save_stores = self.Stores_from_spec(save)
      else:
        save_stores = save
      readonly = not save_stores
    if read is None:
      read_stores = []
    elif isinstance(read, str):
      read_stores = self.Stores_from_spec(read)
    else:
      read_stores = read
    if save2 is None:
      save2_stores = []
    else:
      if isinstance(save2, str):
        save2_stores = self.Stores_from_spec(save2)
      else:
        save2_stores = save2
    if read2 is None:
      read2_stores = []
    elif isinstance(read2, str):
      read2_stores = self.Stores_from_spec(read2)
    else:
      read2_stores = read2
    if copy2 is None:
      copy2_stores = []
    elif isinstance(copy2, str):
      copy2_stores = self.Stores_from_spec(copy2)
    else:
      copy2_stores = copy2
    S = ProxyStore(
        store_name,
        save_stores, read_stores,
        save2=save2_stores, read2=read2_stores,
        copy2=copy2_stores,
    )
    S.readonly = readonly
    return S

  def tcp_Store(
      self,
      store_name,
      *,
      type_=None,
      host=None,
      port=None,
  ):
    ''' Construct a TCPClientStore from a "tcp" clause.
    '''
    if type_ is not None:
      assert type_ == 'tcp'
    if not host:
      host = 'localhost'
    if port is None:
      raise ValueError('no "port"')
    if isinstance(port, str):
      port, _ = get_integer(port, 0)
    return TCPClientStore(store_name, (host, port))

  def socket_Store(
      self,
      store_name,
      *,
      type_=None,
      socket_path=None,
  ):
    ''' Construct a UNIXSocketClientStore from a "socket" clause.
    '''
    if type_ is not None:
      assert type_ == 'socket'
    return UNIXSocketClientStore(store_name, socket_path)



#!/usr/bin/python

''' Assorted debugging assistance functions.
'''

from binascii import hexlify
from cs.fileutils import shortpath
from cs.tty import ttysize
from cs.x import X

def dump_Block(block, indent=''):
  ''' Dump a Block.
  '''
  X("%s%s", indent, block)
  if block.indirect:
    indent += '  '
    subblocks = block.subblocks
    X("%sindirect %d subblocks, span %d bytes",
      indent, len(subblocks), len(block))
    for B in subblocks:
      dump_Block(B, indent=indent)

def dump_Dirent(E, indent='', recurse=False, not_dir=False):
  ''' Dump a Dirent.
  '''
  X("%s%r", indent, E)
  if E.isdir and not not_dir:
    indent += '  '
    for name in sorted(E.keys()):
      E2 = E[name]
      dump_Dirent(E2, indent, recurse=recurse, not_dir=not recurse)

def dump_chunk(data, leadin, max_width=None, one_line=False):
  ''' Dump a data chunk.
  '''
  if max_width is None:
    _, columns = ttysize(1)
    if columns is None:
      columns = 80
    max_width = columns - 1
  leadin += ' %5d' % (len(data),)
  leadin2 = ' ' * len(leadin)
  data_width = max_width - len(leadin)
  slice_size = (data_width - 1) // 3
  assert slice_size > 0
  doff = 0
  while doff < len(data):
    doff2 = doff + slice_size
    chunk = data[doff:doff2]
    hex_text = hexlify(chunk).decode('utf-8')
    txt_text = ''.join(
        c if c.isprintable() else '.'
        for c in chunk.decode('iso8859-1')
    )
    print(leadin, txt_text, hex_text)
    if one_line:
      break
    leadin = leadin2
    doff = doff2

def dump_Store(S, indent=''):
  ''' Dump a description of a Store.
  '''
  from .cache import FileCacheStore
  from .store import MappingStore, ProxyStore, DataDirStore
  X("%s%s:%s", indent, type(S).__name__, S.name)
  indent += '  '
  if isinstance(S, DataDirStore):
    X("%sdir = %s", indent, shortpath(S._datadir.statedirpath))
  elif isinstance(S, FileCacheStore):
    X("%sdatadir = %s", indent, shortpath(S.cache.dirpath))
  elif isinstance(S, ProxyStore):
    for attr in 'save', 'read', 'save2', 'read2', 'copy2':
      backends = getattr(S, attr)
      if backends:
        backends = sorted(backends, key=lambda S: S.name)
        X("%s%s = %s", indent, attr, ','.join(backend.name for backend in backends))
        for backend in backends:
          dump_Store(backend, indent + '  ')
  elif isinstance(S, MappingStore):
    mapping = S.mapping
    X("%smapping = %s", indent, type(mapping))
  else:
    X("%sUNRECOGNISED Store type", indent)



#!/usr/bin/env python
from setuptools import setup
setup(
  name = 'cs.vt',
  description = 'A content hash based data store with a filesystem layer, using variable sized blocks, arbitrarily sized data and utilising some domain knowledge to aid efficient block boundary selection.',
  author = 'Cameron Simpson',
  author_email = 'cs@cskk.id.au',
  version = '20181231.1',
  url = 'https://bitbucket.org/cameron_simpson/css/commits/all',
  classifiers = ['Development Status :: 2 - Pre-Alpha', 'Environment :: Console', 'Programming Language :: Python :: 3', 'Intended Audience :: Developers', 'Operating System :: OS Independent', 'Topic :: Software Development :: Libraries :: Python Modules', 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'],
  entry_points = {'console_scripts': ['vt = cs.vt.__main__:main']},
  install_requires = ['cs.buffer', 'cs.app.flag', 'cs.binary', 'cs.binary_tests', 'cs.cache', 'cs.debug', 'cs.deco', 'cs.env', 'cs.excutils', 'cs.fileutils', 'cs.inttypes', 'cs.later', 'cs.lex', 'cs.logutils', 'cs.mappings', 'cs.packetstream', 'cs.pfx', 'cs.progress', 'cs.py.func', 'cs.py.stack', 'cs.queues', 'cs.range', 'cs.resources', 'cs.result', 'cs.seq', 'cs.serialise', 'cs.socketutils', 'cs.threads', 'cs.tty', 'cs.units', 'cs.x', 'lmdb'],
  keywords = ['python3'],
  long_description = 'A content hash based data store with a filesystem layer, using\nvariable sized blocks, arbitrarily sized data and utilising some\ndomain knowledge to aid efficient block boundary selection.\n\n*NOTE*: pre-Alpha; alpha release following soon once the packaging\nis complete.\n\nSee also the Plan 9 Venti system:\n(http://library.pantek.com/general/plan9.documents/venti/venti.html,\nhttp://en.wikipedia.org/wiki/Venti).\n\n## Class `DebuggingLock`\n\nA wrapper for a threading Lock or RLock\nto notice contention and report contending uses.\n\n## Function `fromtext(s)`\n\nReturn raw byte array from text/hexadecimal string.\n\n## Class `LockContext`\n\nMRO: `builtins.tuple`  \nLockContext(caller, thread)\n\n## Function `totext(data)`\n\nRepresent a byte sequence as a hex/text string.',
  long_description_content_type = 'text/markdown',
  package_dir = {'': 'lib/python'},
  packages = ['cs.vt'],
)



#!/usr/bin/python
#
# Pushpull tests. - Cameron Simpson <cs@cskk.id.au> (formerly cs@zip.com.au) 18dec2015
#

import unittest
from random import randint
from .randutils import rand0, randblock
from cs.x import X
from .hash import HashUtilDict
from .pushpull import missing_hashcodes, missing_hashcodes_by_checksum
from cs.x import X
import cs.x; cs.x.X_via_tty = True

class _TestMissingHashCodes(object):

  def setUp(self):
    self.map1 = HashUtilDict()
    self.map2 = HashUtilDict()

  def test00empty(self):
    ''' Compare two empty maps.
    '''
    missing = list(self.miss_generator(self.map1, self.map2))
    self.assertEqual(len(missing), 0)

  def test01random_identical(self):
    ''' Fill map1 and map2 with identical some random blocks.
    '''
    for _ in range(32):
      data = randblock(rand0(8193))
      h1 = self.map1.add(data)
      h2 = self.map2.add(data)
      self.assertEqual(h1, h2)
    missing = list(self.miss_generator(self.map1, self.map2))
    self.assertEqual(len(missing), 0)

  def test02random1only(self):
    ''' Fill map1 with random blocks, nothing in map2.
    '''
    for n in range(32):
      data = randblock(rand0(8193))
      h1 = self.map1.add(data)
    missing = list(self.miss_generator(self.map1, self.map2))
    self.assertEqual(len(missing), 0)

  def test03random2only(self):
    ''' Fill map2 with random blocks, nothing in map1.
    '''
    ks2 = set()
    for n in range(32):
      data = randblock(rand0(8193))
      h2 = self.map2.add(data)
      ks2.add(h2)
    missing = list(self.miss_generator(self.map1, self.map2))
    self.assertEqual(len(missing), len(ks2))

  def test04random_mixed(self):
    ''' Fill both maps with some overlap.
    '''
    ks1 = set()
    ks2 = set()
    for n in range(32):
      data = randblock(rand0(8193))
      choice = randint(0, 2)
      if choice <= 1:
        h1 = self.map1.add(data)
        ks1.add(h1)
      if choice >= 1:
        h2 = self.map2.add(data)
        ks2.add(h2)
    for window_size in 1, 7, 16, 23, 32, 1024:
      with self.subTest(window_size=window_size):
        # items in map1 not in map2
        missing = set(self.miss_generator(self.map2, self.map1, window_size=window_size))
        self.assertEqual(missing, ks1 - ks2)
        # items in map2 not in map1
        missing = set(self.miss_generator(self.map1, self.map2, window_size=window_size))
        self.assertEqual(missing, ks2 - ks1)

class TestMissingHashCodes_Missing_hashcodes(_TestMissingHashCodes, unittest.TestCase):

  def __init__(self, *a, **kw):
    self.miss_generator = missing_hashcodes
    unittest.TestCase.__init__(self, *a, **kw)

class TestMissingHashCodes_Missing_hashcodes_checksum(_TestMissingHashCodes, unittest.TestCase):

  def __init__(self, *a, **kw):
    self.miss_generator = missing_hashcodes_by_checksum
    unittest.TestCase.__init__(self, *a, **kw)

if __name__ == '__main__':
  from cs.debug import selftest
  selftest('__main__')



#!/usr/bin/python
#
# Hash tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Hash tests.
'''

from hashlib import sha1
from itertools import accumulate
import random
import sys
import unittest
from cs.binary_tests import _TestPacketFields
from .randutils import rand0, randbool, randblock
from . import _TestAdditionsMixin, hash as hash_module
from .hash import Hash_SHA1, decode as decode_hash, HashUtilDict
from .transcribe import Transcriber, parse

class TestDataFilePacketFields(_TestPacketFields, unittest.TestCase):
  ''' Hook to test the hash PacketFields.
  '''

  def setUp(self):
    ''' Test the hash module PacketField classes.
    '''
    self.module = hash_module

class TestHashing(unittest.TestCase):
  ''' Tests for the hashcode facility.
  '''

  def setUp(self):
    ''' Initialise the pseudorandom number generator.
    '''
    random.seed()

  def testSHA1(self):
    ''' Test the SHA1 hash function.
    '''
    for _ in range(10):
      rs = bytes( random.randint(0, 255) for _ in range(100) )
      H = Hash_SHA1.from_chunk(rs)
      self.assertEqual( sha1(rs).digest(), bytes(H) )
      self.assertTrue( isinstance(H, Transcriber))
      Hs = str(H)
      H2, offset = parse(Hs)
      self.assertTrue( offset == len(Hs) )
      self.assertEqual(H, H2)
      # bytes(hash_num + hash_bytes)
      Hencode = H.encode()
      H2, offset = decode_hash(Hencode)
      self.assertEqual(offset, len(Hencode))
      self.assertEqual(H, H2)

class _TestHashCodeUtils(_TestAdditionsMixin, unittest.TestCase):
  ''' Test for the hashcodeutilities.
  '''

  MAP_FACTORY = lambda _: None

  def setUp(self):
    ''' Prepare for testing using the specified map factory.
    '''
    self.maxDiff = 16384
    self.map1 = self.MAP_FACTORY()
    if self.map1 is None:
      raise unittest.SkipTest("no MAP_FACTORY, skipping test")
    self.map1.open()
    self.keys1 = set()
    try:
      keys_method = self.map1.keys
    except AttributeError:
      self.has_keys = False
    else:
      try:
        _ = keys_method()
      except NotImplementedError:
        self.has_keys = False
      else:
        self.has_keys = True

  def tearDown(self):
    ''' Close map1.
    '''
    self.map1.close()

  def test00first(self):
    ''' Trivial test adding 2 blocks.
    '''
    with self.subTest(map_type=type(self.map1)):
      M1 = self.map1
      KS1 = self.keys1
      # test emptiness
      self.assertLen(M1, 0)
      # add one block
      data = randblock(rand0(8193))
      h = M1.add(data)
      KS1.add(h)
      self.assertLen(M1, 1)
      self.assertEqual(set(M1.hashcodes()), KS1)
      # add another block
      data2 = randblock(rand0(8193))
      h2 = M1.add(data2)
      KS1.add(h2)
      self.assertLen(M1, 2)
      self.assertEqual(set(M1.hashcodes()), KS1)

  def test01test_hashcodes_from(self):
    ''' Test the hashcodes_from method.
    '''
    with self.subTest(map_type=type(self.map1)):
      # fill map1 with 16 random data blocks
      M1 = self.map1
      KS1 = self.keys1
      for _ in range(16):
        data = randblock(rand0(8193))
        h = M1.add(data)
        KS1.add(h)
      # make a block not in the map
      data2 = randblock(rand0(8193))
      h2 = Hash_SHA1.from_chunk(data2)
      self.assertNotIn(h2, KS1, "abort test: %s in previous blocks" % (h2,))
      #
      # extract hashes, check results
      #
      ks = sorted(KS1)
      for reverse in False, True:
        for start_hashcode in [None] + ks + [h2]:
          with self.subTest(M1type=type(M1), reverse=reverse, start_hashcode=start_hashcode):
            hs = list(M1.hashcodes_from(start_hashcode=start_hashcode,
                                        reverse=reverse))
            self.assertIsOrdered(hs, reverse=reverse, strict=True)
            if reverse:
              ksrev = reversed(ks)
              hs2 = [ h for h in ksrev if start_hashcode is None or h <= start_hashcode ]
            else:
              hs2 = [ h for h in ks if start_hashcode is None or h >= start_hashcode ]
            hs = list(sorted(hs))
            hs2 = list(sorted(hs2))
            self.assertEqual(hs, hs2)

  def test02hashcodes(self):
    ''' Various tests.
    '''
    with self.subTest(map_type=type(self.map1)):
      M1 = self.map1
      KS1 = self.keys1
      # add 16 random blocks to the map with some sanity checks along the way
      for n in range(16):
        data = randblock(rand0(8193))
        h = M1.add(data)
        self.assertIn(h, M1)
        self.assertNotIn(h, KS1)
        KS1.add(h)
        self.assertIn(h, KS1)
        self.assertLen(M1, n+1)
        self.assertEqual(len(KS1), n+1)
        self.assertEqual(set(iter(M1)), KS1)
        self.assertEqual(set(M1.hashcodes()), KS1)
      # asking for 0 hashcodes is forbidden
      with self.assertRaises(ValueError):
        # NB: using list() to iterate over the generator, thus executing .hashcodes
        hs = list(M1.hashcodes(length=0))
      # fetch the leading n hashcodes from the map, with and without `after`
      for after in False, True:
        with self.subTest(after=after):
          for n in range(1, 16):
            if after:
              start_hashcode = None
              for mincode in accumulate(iter(M1), min):
                start_hashcode = mincode
              if start_hashcode is None:
                # no start_hashcode, skip when after is true
                continue
            else:
              start_hashcode = None
            hs = list(M1.hashcodes(start_hashcode=start_hashcode, after=after, length=n))
            self.assertIsOrdered(hs, False)
            hn = min(n, 15 if after else 16)
            self.assertEqual(len(hs), hn)
      # traverse the map in various sized steps, including random
      sorted_keys = sorted(KS1)
      for step_size in 1, 2, 3, 7, 8, 15, 16, None:
        with self.subTest(step_size=step_size):
          start_hashcode = None
          keys_offset = 0
          seen = set()
          while keys_offset < len(sorted_keys):
            if step_size is None:
              n = random.randint(1, 7)
            else:
              n = step_size
            with self.subTest(
                start_hashcode=start_hashcode,
                keys_offset=keys_offset, n=n,
            ):
              after = start_hashcode is not None
              hs = list(
                  M1.hashcodes(
                      start_hashcode=start_hashcode,
                      length=n,
                      reverse=False,
                      after=after))
              # verify that no key has been seen before
              for h in hs:
                self.assertNotIn(h, seen)
              # verify ordering of returned list
              self.assertIsOrdered(hs, reverse=False, strict=True)
              # verify that least key is > start_hashcode
              if start_hashcode is not None:
                self.assertLess(start_hashcode, hs[0])
              hn = min(len(sorted_keys) - keys_offset, n)
              self.assertEqual(len(hs), hn)
              # verify returned keys against master list
              for i in range(hn):
                self.assertEqual(sorted_keys[keys_offset + i], hs[i])
              # note these keys, advance
              seen.update(hs)
              keys_offset += hn
              start_hashcode = hs[-1]
          # verify that all keys have been retrieved
          self.assertEqual(sorted_keys, sorted(seen))

  def test03hashcodes_missing(self):
    ''' Test the hashcodes_missing function.
    '''
    with self.subTest(map_type=type(self.map1)):
      M1 = self.map1
      KS1 = self.keys1
      for _ in range(16):
        data = randblock(rand0(8193))
        h = M1.add(data)
        KS1.add(h)
      with self.MAP_FACTORY() as M2:
        KS2 = set()
        # construct M2 as a mix of M1 and random new blocks
        for _ in range(16):
          if randbool():
            data = randblock(rand0(8193))
            h = M2.add(data)
            KS2.add(h)
          else:
            M1ks = list(M1.hashcodes())
            M1hash = M1ks[rand0(len(M1ks))]
            data = M1[M1hash]
            h = M2.add(data)
            self.assertEqual(h, M1hash)
            KS2.add(h)
          # compare differences between M1 and M2
          # against differences between key sets KS1 and KS2
          M1missing = sorted(set(M1.hashcodes_missing(M2)))
          KS1missing = sorted(KS2 - KS1)
          self.assertEqual(M1missing, KS1missing)
          M2missing = set(M2.hashcodes_missing(M1))
          KS2missing = KS1 - KS2
          self.assertEqual(M2missing, KS2missing)

class TestHashCodeUtils(_TestHashCodeUtils, unittest.TestCase):
  ''' Test the HashCodeUtils facilities via a HashUitlDict.
  '''

  MAP_FACTORY = HashUtilDict

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Datadir tests.
# - Cameron Simpson <cs@cskk.id.au>
#

import os
from os.path import abspath
import random
import shutil
import sys
import tempfile
import unittest
from .randutils import rand0, randblock
from .datadir import DataDir, DataDirIndexEntry
from .hash import HASHCLASS_BY_NAME
from .index import class_names as indexclass_names, class_by_name as indexclass_by_name

MAX_BLOCK_SIZE = 16383
RUN_SIZE = 100

def mktmpdir(flavour=None):
  ''' Create a temporary scratch directory.
  '''
  prefix = "datadir-test"
  if flavour is not None:
    prefix += '-' + flavour
  return abspath(
      tempfile.mkdtemp(
          prefix="datadir-test",
          suffix=".dir",
          dir='.'))

class TestDataDir(unittest.TestCase):

  ##MAP_FACTORY = lambda self: DataDir(mktmpdir(), mktmpdir(), self.hashclass, self.indexclass)

  def __init__(self, *a, **kw):
    a = list(a)
    method_name = a.pop()
    if a:
      raise ValueError("unexpected arguments: %r" % (a,))
    self.indexdirpath = None
    self.datadirpath = None
    self.indexclass = None
    self.hashclass = None
    self.rollover = None
    self.__dict__.update(kw)
    unittest.TestCase.__init__(self, method_name)

  def setUp(self):
    if self.indexdirpath is None:
      self.indexdirpath = mktmpdir('indexstate')
      self.do_remove_indexdirpath = True
    else:
      self.do_remove_indexdirpath = False
    if self.datadirpath is None:
      self.datadirpath = mktmpdir('data')
      self.do_remove_datadirpath = True
    else:
      self.do_remove_datadirpath = False
    self.datadir = self._open_default_datadir()
    self.datadir.open()
    random.seed()

  def _open_default_datadir(self):
    return DataDir(
        self.indexdirpath,
        self.hashclass,
        indexclass=self.indexclass,
        rollover=self.rollover)

  def tearDown(self):
    self.datadir.close()
    os.system("ls -l -- " + self.datadirpath)
    if self.do_remove_datadirpath:
      shutil.rmtree(self.datadirpath)
    os.system("ls -l -- " + self.indexdirpath)
    if self.do_remove_indexdirpath:
      shutil.rmtree(self.indexdirpath)

  def test000IndexEntry(self):
    ''' Test roundtrip of index entry encode/decode.
    '''
    for _ in range(RUN_SIZE):
      rand_n = random.randint(0, 65536)
      rand_offset = random.randint(0, 65536)
      entry = DataDirIndexEntry(rand_n, rand_offset)
      self.assertEqual(entry.n, rand_n)
      self.assertEqual(entry.offset, rand_offset)
      encoded = entry.encode()
      self.assertIsInstance(encoded, bytes)
      entry2 = DataDirIndexEntry.from_bytes(encoded)
      self.assertEqual(entry, entry2)
      self.assertEqual(entry2.n, rand_n)
      self.assertEqual(entry2.offset, rand_offset)

  def test002randomblocks(self):
    ''' Save random blocks, retrieve in random order.
    '''
    D = self.datadir
    with D:
      hashfunc = D.hashclass.from_chunk
      by_hash = {}
      by_data = {}
      # store RUN_SIZE random blocks
      for n in range(RUN_SIZE):
        with self.subTest(store_block_n=n):
          data = randblock(rand0(MAX_BLOCK_SIZE + 1))
          if data in by_data:
            continue
          hashcode = hashfunc(data)
          # test integrity first
          self.assertFalse(hashcode in by_hash)
          self.assertFalse(data in by_data)
          self.assertFalse(hashcode in D)
          # store block/hashcode
          by_hash[hashcode] = data
          by_data[data] = hashcode
          D[hashcode] = data
          # test integrity afterwards
          self.assertTrue(hashcode in by_hash)
          self.assertTrue(data in by_data)
          self.assertTrue(hashcode in D)
      # now retrieve in random order
      hashcodes = list(by_hash.keys())
      random.shuffle(hashcodes)
      for hashcode in hashcodes:
        with self.subTest(probe_hashcode=hashcode):
          self.assertTrue(hashcode in by_hash)
          self.assertTrue(hashcode in D)
          odata = by_hash[hashcode]
          data = D[hashcode]
          self.assertEqual(data, odata)
    # explicitly close the DataDir and reopen
    # this is because the test framework normally does the outermost open/close
    # and therefore the datadir index lock is still sitting aroung
    D.close()
    D = self.datadir = self._open_default_datadir()
    D.open()
    # reopen the DataDir
    with D:
      hashcodes = list(by_hash.keys())
      random.shuffle(hashcodes)
      for n, hashcode in enumerate(hashcodes):
        with self.subTest(n=n, reprobe_hashcode=hashcode):
          self.assertTrue(hashcode in by_hash)
          self.assertTrue(hashcode in D)
          odata = by_hash[hashcode]
          data = D[hashcode]
          self.assertEqual(data, odata)

def multitest_suite(testcase_class, *a, **kw):
  suite = unittest.TestSuite()
  for method_name in dir(testcase_class):
    if method_name.startswith('test'):
      ta = list(a) + [method_name]
      suite.addTest(testcase_class(*ta, **kw))
  return suite

def selftest(argv):
  suite = unittest.TestSuite()
  for hashname in sorted(HASHCLASS_BY_NAME.keys()):
    hashclass = HASHCLASS_BY_NAME[hashname]
    for indexname in sorted(indexclass_names()):
      indexclass = indexclass_by_name(indexname)
      suite.addTest(multitest_suite(TestDataDir, hashclass=hashclass, indexclass=indexclass))
  runner = unittest.TextTestRunner(failfast=True, verbosity=2)
  runner.run(suite)
  ##if False:
  ##  import cProfile
  ##  cProfile.runctx('unittest.main(__name__, None, argv)', globals(), locals())
  ##else:
  ##  unittest.main(__name__, None, argv)
  ##thread_dump()

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Transcribe tests.
# - Cameron Simpson <cs@cskk.id.au>
#

''' Tests for the text transcription facility.
'''

import sys
import random
import unittest
from uuid import uuid4
from .hash import Hash_SHA1
from .transcribe import _TRANSCRIBE

class TestTranscribe(unittest.TestCase):
  ''' Test the text transcription facility.
  '''

  def setUp(self):
    ''' Seed the random number generator.
    '''
    random.seed()

  def test1(self):
    ''' Basic tests.
    '''
    T = _TRANSCRIBE
    for o in (
        '',
        0, 1, 127, '""', '"abc"', '"de\\f"', '"gh\\\"i"',
        uuid4(),
        Hash_SHA1.from_chunk(
            bytes( random.randint(0, 255) for _ in range(100) )
        ),
    ):
      s = T.transcribe_s(o, None)
      self.assertIsInstance(s, str)
      o2, offset = T.parse(s)
      self.assertEqual(
          offset, len(s),
          "UNPARSED: len(s)=%d, offset=%d" % (len(s), offset))
      self.assertIs(type(o), type(o2))
      ##self.assertEqual(o, o2)

def selftest(argv=None):
  ''' Run unit tests.
  '''
  if argv is None:
    argv = sys.argv
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest()
